from gym.envs.registration import register

from .minedojo_env import MineDojoEnv

environments = [
    ["MinedojoCreative0", "v0"],
    ["MinedojoCreative1", "v0"],
    ["MinedojoCreative2", "v0"],
    ["MinedojoCreative4", "v0"],
    ["MinedojoCreative5", "v0"],
    ["MinedojoCreative7", "v0"],
    ["MinedojoCreative8", "v0"],
    ["MinedojoCreative9", "v0"],
]

for environment in environments:
    register(
        id="{}-{}".format(environment[0], environment[1]),
        entry_point="smartplay.minedojo:MineDojoEnv",
        kwargs={"task_id": environment[0][-1]},
    )



from abc import ABC, abstractmethod


class BaseProvider(ABC):
    @abstractmethod
    async def async_chat_completion(self, messages, model, temperature, max_tokens):
        pass

    @abstractmethod
    def sync_chat_completion(self, messages, model, temperature, max_tokens):
        pass



import pandas as pd
import numpy as np

# Constants
COLUMNS = ["column1", "column2", "column3", "column4", "column5"]


def task_func(df, dct):
    if not isinstance(df, pd.DataFrame):
        raise ValueError("The input df is not a DataFrame")
    # Replace values using dictionary mapping
    df = df.replace(dct)

    # Calculate the correlation matrix
    correlation_matrix = np.corrcoef(df.values, rowvar=False)

    return pd.DataFrame(correlation_matrix, columns=df.columns, index=df.columns)


import unittest
import pandas as pd


class TestCases(unittest.TestCase):
    def test_case_1(self):
        # Test with simple numeric DataFrame
        df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
        dct = {1: 10, 2: 20, 3: 30, 4: 40, 5: 50, 6: 60}
        result = task_func(df, dct)
        self.assertTrue(result.shape == (2, 2))

    def test_case_2(self):
        # Test with DataFrame containing NaN values
        df = pd.DataFrame({"A": [1, 2, None], "B": [4, None, 6]})
        dct = {1: 10, 2: 20, 4: 40, 6: 60}
        result = task_func(df, dct)
        self.assertTrue(result.isna().sum().sum() > 0)

    def test_case_3(self):
        # Test with DataFrame containing negative values
        df = pd.DataFrame({"A": [-1, -2, -3], "B": [-4, -5, -6]})
        dct = {-1: 1, -2: 2, -3: 3, -4: 4, -5: 5, -6: 6}
        result = task_func(df, dct)
        self.assertTrue(result.shape == (2, 2))

    def test_case_4(self):
        # Test with DataFrame containing mixed data types
        df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
        dct = {1: 0, 2: 1, 3: 2, 4: 3, 5: 4, 6: 5}
        result = task_func(df, dct)
        self.assertTrue(result.shape == (2, 2))

    def test_case_5(self):
        # Test with larger DataFrame
        df = pd.DataFrame({"A": range(10), "B": range(10, 20), "C": range(20, 30)})
        dct = {i: i + 1 for i in range(30)}
        result = task_func(df, dct)
        self.assertTrue(result.shape == (3, 3))

    def test_case_6(self):
        with self.assertRaises(ValueError):
            task_func("non_df", {})



import json
import traceback
from typing import Optional

import loguru

# from dotenv import load_dotenv
from pydantic import BaseModel

from apropos.src.core.lms.vendors.groq_api import GroqAPIProvider
from apropos.src.core.lms.vendors.openai_api import OpenAIAPIProvider
from typing import Literal

logger = loguru.logger


class JsonDebuggerResponse(BaseModel):
    what_the_error_is: str
    fixed_json_string: str

def get_provider_and_model(provider: Literal["groq", "openai"]):
    if provider == "groq":
        return GroqAPIProvider(force_structured_output=True), "llama3-70b-8192"
    elif provider == "openai":
        return OpenAIAPIProvider(force_structured_output=True), "gpt-4o-mini"
    else:
        raise ValueError("Invalid provider")

def get_messages(
    error_log: list[str],
    string_trying_to_be_parsed_as_json: str,
    response_model: Optional[BaseModel],
):
    response_model_snippet = ""
    if response_model:
        response_model_snippet = f"\nThis json should adhere to the following standard derived from a Pydantic BaseModel: \n\n{response_model.schema()}.\n If you believe the provided response does not contain enough information to be reformatted into the correct JSON, do not make up information, but instead include ESCALATE in your response."
    system_message = f"""
# Instructions
You will be given a JSON string that is not properly formatted, together with the traceback generated by invoking json.loads on the string. Your task is to identify the error in the JSON string and correct it so that it can be parsed successfully.
{response_model_snippet}
Respond ONLY with the corrected JSON string.

# Formatting
Format your response like:
```json
$YOUR_CORRECTED_JSON
```

"""
    user_message = f"""
# Error Log (Last is most recent)
{error_log[-5:]}

# JSON String
{string_trying_to_be_parsed_as_json}

Your correction: """
    messages = [
        {"role": "system", "content": system_message},
        {"role": "user", "content": user_message},
    ]  
    return messages

def parse_outputs(
    fixed_json_string_raw: str,
):
    if "ESCALATE" in fixed_json_string_raw:
        return "ESCALATE"
    if "```json" not in fixed_json_string_raw and "```" in fixed_json_string_raw:
        fixed_json_string_raw = fixed_json_string_raw.split("```")[1].split("```")[
            0
        ]
    elif (
        "```json" not in fixed_json_string_raw
        and "```" not in fixed_json_string_raw
    ):
        fixed_json_string = fixed_json_string_raw
    else:
        fixed_json_string = fixed_json_string_raw.split("```json")[1].split("```")[
            0
        ]
    return fixed_json_string
# TODO: add support for local models?
async def json_debugger_async(
    string_trying_to_be_parsed_as_json: str,
    response_model: Optional[BaseModel],
    provider_name: Literal["groq", "openai"] = "openai",
):
    provider, model = get_provider_and_model(provider_name)
    succeeded = False
    valid_json = None
    patience = 6
    error_log = []

    try:
        json.loads(string_trying_to_be_parsed_as_json)
        succeeded = True
        return json.loads(string_trying_to_be_parsed_as_json)
    except:
        error_log.append("Error in healing: " + traceback.format_exc())
        patience -= 1
    while not succeeded:
        messages = get_messages(error_log, string_trying_to_be_parsed_as_json, response_model)
        fixed_json_string_raw = await provider.async_chat_completion(
            messages=messages,
            model=model,
            temperature=0.0,
            max_tokens=1000,
        )
        fixed_json_string = parse_outputs(fixed_json_string_raw)
        try:
            json.loads(fixed_json_string)
            succeeded = True
            valid_json = fixed_json_string
        except Exception as e:
            error_log.append(
                "Attempted fix: \n\n"
                + fixed_json_string
                + "\n\nError: "
                + traceback.format_exc()
            )
            patience -= 1
        if patience == 0:
            break
    assert succeeded, "Failed to fix JSON string"
    assert not valid_json is None, "Failed to fix JSON string"
    return json.loads(valid_json)


def json_debugger_sync(
    string_trying_to_be_parsed_as_json: str,
    response_model: Optional[BaseModel],
    provider_name: Literal["groq", "openai"] = "openai",
):
    provider, model = get_provider_and_model(provider_name)
    succeeded = False
    valid_json = None
    patience = 6
    error_log = []

    try:
        json.loads(string_trying_to_be_parsed_as_json)
        succeeded = True
        return json.loads(string_trying_to_be_parsed_as_json)
    except:
        error_log.append("Error in healing: " + traceback.format_exc())
        patience -= 1
    while not succeeded:
        messages = get_messages(error_log, string_trying_to_be_parsed_as_json, response_model)
        fixed_json_string_raw = provider.sync_chat_completion(
            messages=messages,
            model=model,
            temperature=0.0,
            max_tokens=1000,
        )
        fixed_json_string = parse_outputs(fixed_json_string_raw)
        try:
            json.loads(fixed_json_string)
            succeeded = True
            valid_json = fixed_json_string
        except Exception as e:
            error_log.append(
                "Attempted fix: \n\n"
                + fixed_json_string
                + "\n\nError: "
                + traceback.format_exc()
            )
            patience -= 1
        if patience == 0:
            break
    assert succeeded, "Failed to fix JSON string"
    assert not valid_json is None, "Failed to fix JSON string"
    return json.loads(valid_json)



import io
from typing import List
import modal
from modal import App, Sandbox
from modal.exception import SandboxTimeoutError
from apropos.src.bench.bigcodebench.backends.shared import get_imports


app = App("bigcodebench")


async def execute_code_remotely_modal(
    question_dict, code_solution
):  # , packages: List[str] = []
    solution = question_dict["eval_info"]["code_prompt"] + code_solution
    packages, _ = get_imports(solution)
    test = question_dict["eval_info"]["test"]
    complete = solution + "\n" + test

    unit_test_script = """
import unittest
import io
def test_code():
    path = "script.py"
    loader = unittest.TestLoader()
    suite = loader.discover('.', pattern=path)
    runner = unittest.TextTestRunner()
    result = runner.run(suite)

    # Convert result to a serializable format
    result_dict = {
        "errors": len(result.errors),
        "failures": len(result.failures),
        "testsRun": result.testsRun,
        "wasSuccessful": result.wasSuccessful()
    }
    return result.wasSuccessful(), result_dict

if __name__ == "__main__":
    success,result = test_code()
    print("Success:", success)
    print(result)
"""
    # install_command = ""
    # if packages:
    #     install_command = f"pip install {' '.join(packages)} && "

    # execution_image = modal.Image.from_registry("python:3.9-slim").pip_install(
    #     ""
    #     # "pandas==1.3.3",
    #     # "fuzzywuzzy==0.18.0",
    #     # "openai==1.3.2",
    #     # "ruptures==1.1.8",
    #     # "tiktoken==0.4.0",
    #     # "matplotlib==3.8.1",
    #     # "tabulate==0.9.0",
    #     # "PyYAML",
    # )
    image = modal.Image.debian_slim(python_version="3.9")
    if packages:
        image = image.pip_install("uv")
        package_install_command = " ".join(packages)
        image = image.run_commands(
            f"uv pip install --system --compile-bytecode {package_install_command}"
        )
    try:
        with modal.NetworkFileSystem.ephemeral() as nfs:
            await nfs.write_file.aio("script.py", io.BytesIO(complete.encode()))
            await nfs.write_file.aio(
                "unit_test_script.py", io.BytesIO(unit_test_script.encode())
            )
            sb = Sandbox.create(
                "bash",
                "-c",
                f"{install_command}cd /vol && python -W ignore unit_test_script.py",
                image=execution_image,
                timeout=60,
                cloud="aws",
                network_file_systems={"/vol": nfs},
            )
            await sb.wait.aio()
            stdout = await sb.stdout.read.aio()
            stderr = await sb.stderr.read.aio()
            print("Stdout: ", stdout)
            print("Stderr: ", stderr)
            success = stdout.split("Success: ")[1].split("\n")[0] == "True"
            result_dict = eval(stdout.split("Success: ")[1].split("\n")[1])
            return success, result_dict
    except SandboxTimeoutError:
        return False, {
            "errors": 1,
            "failures": 0,
            "testsRun": 0,
            "wasSuccessful": False,
            "timeout": True,
        }


if __name__ == "__main__":
    # test
    import asyncio
    from apropos.src.bench.bigcodebench.main import BigCodeBenchComplete_Benchmark

    question = BigCodeBenchComplete_Benchmark().train[0]
    print(
        asyncio.run(
            execute_code_remotely_modal(
                question.information, question.information["answer"], ["pandas"]
            )
        )
    )



import setuptools
import pathlib


setuptools.setup(
    name="smartplay",
    version="0.0.1",
    description="A benchmarking tool for LLMs with games",
    url="",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_namespace_packages("src"),
    package_dir={"": "src"},
    # package_data={'crafter': ['data.yaml', 'assets/*']},
    entry_points={"console_scripts": ["smartplay=smartplay.run_gui:main"]},
    install_requires=[
        "numpy",
        "pandas",
        "pygame",
        "connected-components-3d",
        "gym",
        "minedojo",
        "imageio",
        "pillow",
        "opensimplex",
        "ruamel.yaml<0.18",  # safe_load removed in 0.18
        "importlib-metadata==6.6.0",
        "importlib-resources==5.12.0",
        "vgdl @ git+https://github.com/ahjwang/py-vgdl",
        "setuptools",  # pkg-resources required by vgdl
    ],
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Games/Entertainment",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import collections

import numpy as np

from . import constants
from . import engine
from . import objects
from . import worldgen


# Gym is an optional dependency.
try:
    import gym

    DiscreteSpace = gym.spaces.Discrete
    BoxSpace = gym.spaces.Box
    DictSpace = gym.spaces.Dict
    BaseClass = gym.Env
except ImportError:
    DiscreteSpace = collections.namedtuple("DiscreteSpace", "n")
    BoxSpace = collections.namedtuple("BoxSpace", "low, high, shape, dtype")
    DictSpace = collections.namedtuple("DictSpace", "spaces")
    BaseClass = object


class Env(BaseClass):
    def __init__(
        self,
        area=(64, 64),
        view=(9, 9),
        size=(64, 64),
        reward=True,
        length=10000,
        seed=None,
    ):
        view = np.array(view if hasattr(view, "__len__") else (view, view))
        size = np.array(size if hasattr(size, "__len__") else (size, size))
        seed = np.random.randint(0, 2**31 - 1) if seed is None else seed
        self._area = area
        self._view = view
        self._size = size
        self._reward = reward
        self._length = length
        self._seed = seed
        self._episode = 0
        self._world = engine.World(area, constants.materials, (12, 12))
        self._textures = engine.Textures(constants.root / "assets")
        item_rows = int(np.ceil(len(constants.items) / view[0]))
        self._local_view = engine.LocalView(
            self._world, self._textures, [view[0], view[1] - item_rows]
        )
        self._item_view = engine.ItemView(self._textures, [view[0], item_rows])
        self._sem_view = engine.SemanticView(
            self._world,
            [
                objects.Player,
                objects.Cow,
                objects.Zombie,
                objects.Skeleton,
                objects.Arrow,
                objects.Plant,
            ],
        )
        self._step = None
        self._player = None
        self._last_health = None
        self._unlocked = None
        # Some libraries expect these attributes to be set.
        self.reward_range = None
        self.metadata = None

    @property
    def observation_space(self):
        return BoxSpace(0, 255, tuple(self._size) + (3,), np.uint8)

    @property
    def action_space(self):
        return DiscreteSpace(len(constants.actions))

    @property
    def action_names(self):
        return constants.actions

    def reset(self):
        center = (self._world.area[0] // 2, self._world.area[1] // 2)
        self._episode += 1
        self._step = 0
        self._world.reset(seed=hash((self._seed, self._episode)) % (2**31 - 1))
        self._update_time()
        self._player = objects.Player(self._world, center)
        self._last_health = self._player.health
        self._world.add(self._player)
        self._unlocked = set()
        worldgen.generate_world(self._world, self._player)
        return self._obs()

    def step(self, action):
        self._step += 1
        self._update_time()
        self._player.action = constants.actions[action]
        for obj in self._world.objects:
            if self._player.distance(obj) < 2 * max(self._view):
                obj.update()
        if self._step % 10 == 0:
            for chunk, objs in self._world.chunks.items():
                # xmin, xmax, ymin, ymax = chunk
                # center = (xmax - xmin) // 2, (ymax - ymin) // 2
                # if self._player.distance(center) < 4 * max(self._view):
                self._balance_chunk(chunk, objs)
        obs = self._obs()
        reward = (self._player.health - self._last_health) / 10
        self._last_health = self._player.health
        unlocked = {
            name
            for name, count in self._player.achievements.items()
            if count > 0 and name not in self._unlocked
        }
        if unlocked:
            self._unlocked |= unlocked
            reward += 1.0
        dead = self._player.health <= 0
        over = self._length and self._step >= self._length
        done = dead or over
        info = {
            "inventory": self._player.inventory.copy(),
            "achievements": self._player.achievements.copy(),
            "sleeping": self._player.sleeping,
            "discount": 1 - float(dead),
            "semantic": self._sem_view(),
            "player_pos": self._player.pos,
            "player_facing": self._player.facing,
            "reward": reward,
            "dead": dead,
            "unlocked": unlocked,
            "action": self._player.action,
            "view": self._view,
        }
        if not self._reward:
            reward = 0.0
        return obs, reward, done, info

    def render(self, size=None):
        size = size or self._size
        unit = size // self._view
        canvas = np.zeros(tuple(size) + (3,), np.uint8)
        local_view = self._local_view(self._player, unit)
        item_view = self._item_view(self._player.inventory, unit)
        view = np.concatenate([local_view, item_view], 1)
        border = (size - (size // self._view) * self._view) // 2
        (x, y), (w, h) = border, view.shape[:2]
        canvas[x : x + w, y : y + h] = view
        return canvas.transpose((1, 0, 2))

    def _obs(self):
        return self.render()

    def _update_time(self):
        # https://www.desmos.com/calculator/grfbc6rs3h
        progress = (self._step / 300) % 1 + 0.3
        daylight = 1 - np.abs(np.cos(np.pi * progress)) ** 3
        self._world.daylight = daylight

    def _balance_chunk(self, chunk, objs):
        light = self._world.daylight
        self._balance_object(
            chunk,
            objs,
            objects.Zombie,
            "grass",
            6,
            0,
            0.3,
            0.4,
            lambda pos: objects.Zombie(self._world, pos, self._player),
            lambda num, space: (0 if space < 50 else 3.5 - 3 * light, 3.5 - 3 * light),
        )
        self._balance_object(
            chunk,
            objs,
            objects.Skeleton,
            "path",
            7,
            7,
            0.1,
            0.1,
            lambda pos: objects.Skeleton(self._world, pos, self._player),
            lambda num, space: (0 if space < 6 else 1, 2),
        )
        self._balance_object(
            chunk,
            objs,
            objects.Cow,
            "grass",
            5,
            5,
            0.01,
            0.1,
            lambda pos: objects.Cow(self._world, pos),
            lambda num, space: (0 if space < 30 else 1, 1.5 + light),
        )

    def _balance_object(
        self,
        chunk,
        objs,
        cls,
        material,
        span_dist,
        despan_dist,
        spawn_prob,
        despawn_prob,
        ctor,
        target_fn,
    ):
        xmin, xmax, ymin, ymax = chunk
        random = self._world.random
        creatures = [obj for obj in objs if isinstance(obj, cls)]
        mask = self._world.mask(*chunk, material)
        target_min, target_max = target_fn(len(creatures), mask.sum())
        if len(creatures) < int(target_min) and random.uniform() < spawn_prob:
            xs = np.tile(np.arange(xmin, xmax)[:, None], [1, ymax - ymin])
            ys = np.tile(np.arange(ymin, ymax)[None, :], [xmax - xmin, 1])
            xs, ys = xs[mask], ys[mask]
            i = random.randint(0, len(xs))
            pos = np.array((xs[i], ys[i]))
            empty = self._world[pos][1] is None
            away = self._player.distance(pos) >= span_dist
            if empty and away:
                self._world.add(ctor(pos))
        elif len(creatures) > int(target_max) and random.uniform() < despawn_prob:
            obj = creatures[random.randint(0, len(creatures))]
            away = self._player.distance(obj.pos) >= despan_dist
            if away:
                self._world.remove(obj)


class EnvSample(BaseClass):
    def __init__(
        self,
        area=(64, 64),
        view=(9, 9),
        size=(64, 64),
        reward=True,
        length=10000,
        seed=None,
        steps=20,
    ):
        view = np.array(view if hasattr(view, "__len__") else (view, view))
        size = np.array(size if hasattr(size, "__len__") else (size, size))
        seed = np.random.randint(0, 2**31 - 1) if seed is None else seed
        self._area = area
        self._view = view
        self._size = size
        self._reward = reward
        self._length = length
        self._seed = seed
        self._episode = 0
        self._world = engine.World(area, constants.materials, (12, 12))
        self._textures = engine.Textures(constants.root / "assets")
        item_rows = int(np.ceil(len(constants.items) / view[0]))
        self._local_view = engine.LocalView(
            self._world, self._textures, [view[0], view[1] - item_rows]
        )
        self._item_view = engine.ItemView(self._textures, [view[0], item_rows])
        self._sem_view = engine.SemanticView(
            self._world,
            [
                objects.Player,
                objects.Cow,
                objects.Zombie,
                objects.Skeleton,
                objects.Arrow,
                objects.Plant,
            ],
        )
        self._step = None
        self._nsteps = steps
        self._player = None
        self._last_health = None
        self._unlocked = None
        # Some libraries expect these attributes to be set.
        self.reward_range = None
        self.metadata = None

    @property
    def observation_space(self):
        return BoxSpace(0, 255, tuple(self._size) + (3,), np.uint8)

    @property
    def action_space(self):
        return DiscreteSpace(len(constants.actions))

    @property
    def action_names(self):
        return constants.actions

    def reset(self):
        center = (self._world.area[0] // 2, self._world.area[1] // 2)
        self._episode += 1
        self._step = 0
        self._world.reset(seed=hash((self._seed, self._episode)) % (2**31 - 1))
        self._update_time()
        self._player = objects.Player(self._world, center)
        self._last_health = self._player.health
        self._world.add(self._player)
        self._unlocked = set()
        worldgen.generate_world(self._world, self._player)
        return self._obs()

    def __step(self, action):
        self._step += 1
        self._update_time()
        self._player.action = constants.actions[action]
        for obj in self._world.objects:
            if self._player.distance(obj) < 2 * max(self._view):
                obj.update()
        if self._step % 10 == 0:
            for chunk, objs in self._world.chunks.items():
                # xmin, xmax, ymin, ymax = chunk
                # center = (xmax - xmin) // 2, (ymax - ymin) // 2
                # if self._player.distance(center) < 4 * max(self._view):
                self._balance_chunk(chunk, objs)
        obs = self._obs()
        reward = (self._player.health - self._last_health) / 10
        self._last_health = self._player.health
        unlocked = {
            name
            for name, count in self._player.achievements.items()
            if count > 0 and name not in self._unlocked
        }
        if unlocked:
            self._unlocked |= unlocked
            reward += 1.0
        dead = self._player.health <= 0
        over = self._length and self._step >= self._length
        done = dead or over
        info = {
            "inventory": self._player.inventory.copy(),
            "achievements": self._player.achievements.copy(),
            "discount": 1 - float(dead),
            "semantic": self._sem_view(),
            "player_pos": self._player.pos,
            "player_facing": self._player.facing,
            "reward": reward,
            "dead": dead,
            "unlocked": unlocked,
            "action": self._player.action,
            "view": self._view,
        }
        if not self._reward:
            reward = 0.0
        return obs, reward, done, info

    def step(self, action):
        frames = max(int(np.random.normal(loc=self._nsteps, scale=self._nsteps / 2)), 1)
        for _ in range(frames):
            obs, reward, done, info = self.__step(
                np.random.randint(len(constants.actions))
            )
            if done:
                self.reset()
                obs, reward, done, info = self.__step(
                    np.random.randint(len(constants.actions))
                )
        return obs, reward, done, info

    def render(self, size=None):
        size = size or self._size
        unit = size // self._view
        canvas = np.zeros(tuple(size) + (3,), np.uint8)
        local_view = self._local_view(self._player, unit)
        item_view = self._item_view(self._player.inventory, unit)
        view = np.concatenate([local_view, item_view], 1)
        border = (size - (size // self._view) * self._view) // 2
        (x, y), (w, h) = border, view.shape[:2]
        canvas[x : x + w, y : y + h] = view
        return canvas.transpose((1, 0, 2))

    def _obs(self):
        return self.render()

    def _update_time(self):
        # https://www.desmos.com/calculator/grfbc6rs3h
        progress = (self._step / 300) % 1 + 0.3
        daylight = 1 - np.abs(np.cos(np.pi * progress)) ** 3
        self._world.daylight = daylight

    def _balance_chunk(self, chunk, objs):
        light = self._world.daylight
        self._balance_object(
            chunk,
            objs,
            objects.Zombie,
            "grass",
            6,
            0,
            0.3,
            0.4,
            lambda pos: objects.Zombie(self._world, pos, self._player),
            lambda num, space: (0 if space < 50 else 3.5 - 3 * light, 3.5 - 3 * light),
        )
        self._balance_object(
            chunk,
            objs,
            objects.Skeleton,
            "path",
            7,
            7,
            0.1,
            0.1,
            lambda pos: objects.Skeleton(self._world, pos, self._player),
            lambda num, space: (0 if space < 6 else 1, 2),
        )
        self._balance_object(
            chunk,
            objs,
            objects.Cow,
            "grass",
            5,
            5,
            0.01,
            0.1,
            lambda pos: objects.Cow(self._world, pos),
            lambda num, space: (0 if space < 30 else 1, 1.5 + light),
        )

    def _balance_object(
        self,
        chunk,
        objs,
        cls,
        material,
        span_dist,
        despan_dist,
        spawn_prob,
        despawn_prob,
        ctor,
        target_fn,
    ):
        xmin, xmax, ymin, ymax = chunk
        random = self._world.random
        creatures = [obj for obj in objs if isinstance(obj, cls)]
        mask = self._world.mask(*chunk, material)
        target_min, target_max = target_fn(len(creatures), mask.sum())
        if len(creatures) < int(target_min) and random.uniform() < spawn_prob:
            xs = np.tile(np.arange(xmin, xmax)[:, None], [1, ymax - ymin])
            ys = np.tile(np.arange(ymin, ymax)[None, :], [xmax - xmin, 1])
            xs, ys = xs[mask], ys[mask]
            i = random.randint(0, len(xs))
            pos = np.array((xs[i], ys[i]))
            empty = self._world[pos][1] is None
            away = self._player.distance(pos) >= span_dist
            if empty and away:
                self._world.add(ctor(pos))
        elif len(creatures) > int(target_max) and random.uniform() < despawn_prob:
            obj = creatures[random.randint(0, len(creatures))]
            away = self._player.distance(obj.pos) >= despan_dist
            if away:
                self._world.remove(obj)



import copy
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Tuple, Type

import networkx as nx
from pydantic import BaseModel, create_model

from apropos.src.core.programs.prompt import Demonstration


def get_python_type(json_type: str, item_type: str = None) -> Type:
    """
    Maps JSON schema types to Python types.

    Args:
    json_type (str): The JSON schema type.
    item_type (str): The type of items in an array, if applicable.

    Returns:
    Type: The corresponding Python type.
    """
    type_mapping = {
        "string": str,
        "integer": int,
        "number": float,
        "boolean": bool,
        "array": List[get_python_type(item_type)] if item_type else List,
    }
    return type_mapping.get(json_type, str)


def create_pydantic_model_from_schema(
    schema: Dict[str, Any], model_name: str = "DynamicModel"
) -> BaseModel:
    """
    Creates a Pydantic model dynamically from a given schema.

    Args:
    schema (Dict[str, Any]): The schema dictionary of the model.
    model_name (str): The name of the dynamically created model.

    Returns:
    BaseModel: A dynamically created Pydantic model class.
    """
    field_definitions = {}
    for k, v in schema["properties"].items():
        field_type = v.get("type", "string")
        if field_type == "array":
            item_type = v["items"]["type"]
            python_type = get_python_type(field_type, item_type)
        else:
            python_type = get_python_type(field_type)
        field_definitions[k] = (python_type, ...)
    return create_model(model_name, **field_definitions)


@dataclass
class PromptRecord:
    user_message: str
    system_message: str
    outputs: Dict
    lm_name: str


@dataclass
class StageRecord:
    name: str
    inputs: Dict
    outputs: Dict
    prompt_record: PromptRecord

    def to_dict(self):
        return {
            "name": self.name,
            "inputs": self.inputs,
            "outputs": self.outputs,
            "prompt_record": {
                "user_message": self.prompt_record.user_message,
                "system_message": self.prompt_record.system_message,
                "outputs": self.prompt_record.outputs,
                "lm_name": self.prompt_record.lm_name,
            },
        }


@dataclass
class AnnotatedStageRecord(StageRecord):
    name: str
    inputs: Dict
    outputs: Dict
    prompt_record: PromptRecord
    efficacious: bool
    annotation: str

    def to_demonstration(self) -> Demonstration:
        return Demonstration(
            inputs=self.inputs,
            outputs=self.outputs,
            prompt=self.prompt_record.system_message,
            gold_outputs=None,
            annotation=self.annotation,
        )


@dataclass
class DagRecord:
    stage_records: List[StageRecord]
    name_graph: nx.DiGraph

    def trivially_annotate(self):
        annotated_stages = []
        for stage in self.stage_records:
            annotated_stages.append(
                AnnotatedStageRecord(
                    name=stage.name,
                    inputs=stage.inputs,
                    outputs=stage.outputs,
                    prompt_record=stage.prompt_record,
                    efficacious=None,
                    annotation=None,
                )
            )
        return AnnotatedDagRecord(
            stage_records=annotated_stages,
            name_graph=self.name_graph,
            efficacious=None,
            annotation=None,
        )

    def to_dict(self):
        return {
            "stage_records": [stage.to_dict() for stage in self.stage_records],
            "name_graph": nx.to_dict_of_dicts(self.name_graph),
        }


@dataclass
class AnnotatedDagRecord(DagRecord):
    stage_records: List[AnnotatedStageRecord]
    name_graph: nx.DiGraph
    efficacious: bool
    annotation: str

    def to_stage_demonstrations(
        self,
        include_annotation=False,
        include_efficacy=False,
        include_inputs=True,
        include_outputs=True,
        include_prompt=False,
    ) -> List[Dict]:
        demonstrations = []
        for stage_record in self.stage_records:
            demonstrations.append(stage_record.to_demonstration())
        return demonstrations

    def to_dag_demonstration(
        self,
        include_annotation=False,
        include_efficacy=False,
        include_inputs=True,
        include_outputs=True,
        include_prompt=False,
    ) -> str:
        demonstration = ""
        for i, stage_record in enumerate(self.stage_records):
            demonstration += f"# Stage {i+1}\n"
            demonstration += stage_record.to_demonstration().to_string()
        if include_annotation:
            demonstration += f"Annotation for program result: {self.annotation}\n"
        if include_efficacy:
            demonstration += f"Efficacy: {self.efficacious}\n"
        return demonstration


class Runnable:
    @abstractmethod
    async def arun(self):
        pass

    @abstractmethod
    def run(self):
        pass

    def to_dict(self):
        d = {}
        if hasattr(self, "llm_config"):
            d["llm_config"] = self.llm_config
        if hasattr(self, "prompt"):
            d["prompt"] = self.prompt.to_dict()
        return d


@dataclass
class TextGraphNode:
    name: str
    inputs: Dict[str, Any]
    outputs: Dict[str, Any]
    get_input_fields: Runnable
    transform: Runnable
    produce_stage_record: Runnable

    def to_dict(self):
        return {
            "name": self.name,
            "runnables": {
                "get_input_fields": self.get_input_fields.to_dict(),
                "transform": self.transform.to_dict(),
                "produce_stage_record": self.produce_stage_record.to_dict(),
            },
        }

    async def arun(self, inputs: Dict[str, Any], verbose: bool = False):
        self.inputs = inputs
        self.outputs = await self.transform.arun(inputs)
        return self.outputs

    def run(self, inputs: Dict[str, Any], verbose: bool = False):
        self.inputs = inputs
        self.outputs = self.transform.run(inputs)
        return self.outputs

    async def get_record(self):
        return await self.produce_stage_record.arun(
            inputs=self.inputs, outputs=self.outputs
        )

    def get_record_sync(self):
        return self.produce_stage_record.run(inputs=self.inputs, outputs=self.outputs)

    def get_inputs(self):
        return self.get_input_fields.run()


@dataclass
class LM_DAG:
    nodes: Dict[str, TextGraphNode]
    edges: List[Tuple[Tuple[str, str], Tuple[str, str]]]
    input_aliases: Dict[str, str]
    output_aliases: Dict[str, str]

    def __init__(self, nodes, edges, input_aliases={}, output_aliases={}):
        self.nodes = {node.name: node for node in nodes}
        self.edges = edges
        self.input_aliases = input_aliases
        self.output_aliases = output_aliases

    def to_dict(self):
        return {
            "nodes": {
                node_name: node.to_dict() for node_name, node in self.nodes.items()
            },
            "edges": self.edges,
            "input_aliases": self.input_aliases,
            "output_aliases": self.output_aliases,
        }

    def get_node_ancestors(self, node_name):
        graph = nx.DiGraph()
        for edge in self.edges:
            graph.add_edge(edge[0][0], edge[1][0])
        # Retrieve all ancestors of the node
        ancestors = nx.ancestors(graph, node_name)
        return list(nx.topological_sort(graph.subgraph(ancestors)))

    # update names

    def get_node_children(self, node_name):
        graph = nx.DiGraph()
        for edge in self.edges:
            graph.add_edge(edge[0][0], edge[1][0])
        children = [n for n in graph.successors(node_name)]
        return children

    def get_node_descendants(self, node_name):
        graph = nx.DiGraph()
        for edge in self.edges:
            graph.add_edge(edge[0][0], edge[1][0])
        descendants = nx.descendants(graph, node_name)
        return list(nx.topological_sort(graph.subgraph(descendants)))

    def get_node_relationship(self, node_a, node_b):
        graph = nx.DiGraph()
        for edge in self.edges:
            graph.add_edge(edge[0][0], edge[1][0])

        if nx.has_path(graph, node_a, node_b):
            return "parent"
        elif nx.has_path(graph, node_b, node_a):
            return "child"
        else:
            return "neither"

    def compile(self):
        g = nx.DiGraph()
        for edge in self.edges:
            g.add_edge(edge[0][0], edge[1][0])
        return g

    async def arun(
        self,
        base_inputs: Dict[str, Any],
        frozen_params: Dict = {"dag_record": None, "unfrozen_node_names": []},
        verbose: bool = False,
    ):
        if frozen_params["dag_record"] is not None:
            return await self.arun_frozen(
                base_inputs,
                frozen_params["dag_record"],
                frozen_params["unfrozen_node_names"],
                verbose=verbose,
            )
        else:
            return await self.arun_standard(base_inputs, verbose)

    async def arun_frozen(
        self,
        base_inputs: Dict[str, Any],
        dag_record: DagRecord,
        unfrozen_node_names: List[str],
        verbose: bool = False,
    ):
        all_unfrozen_node_names = set(unfrozen_node_names)
        for node_name in unfrozen_node_names:
            all_unfrozen_node_names.update(self.get_node_descendants(node_name))
        frozen_node_names = set(self.nodes.keys()) - all_unfrozen_node_names
        unaliased_base_inputs = copy.deepcopy(base_inputs)
        for k, v in unaliased_base_inputs.items():
            if k in self.input_aliases:
                del base_inputs[k]
                base_inputs[self.input_aliases[k]] = v

        assert set(
            [edge[0][1] for edge in self.edges if edge[0][0] == "DAG_INPUT"]
        ).issubset(
            set(base_inputs.keys())
        ), f"Inputs needed: {[edge[0][1] for edge in self.edges if edge[0][0] == 'DAG_INPUT']}, Inputs given: {base_inputs.keys()}"
        stage_records = []
        g = self.compile()
        outputs = {}
        for node_name in nx.topological_sort(g):
            if node_name in ["DAG_INPUT", "DAG_OUTPUT"]:
                continue
            node = self.nodes[node_name]
            inputs_needed = node.get_inputs()
            inputs = {}
            for edge in self.edges:
                if edge[1][0] != node_name:
                    continue
                if edge[1][0] == "DAG_OUTPUT":
                    continue
                if edge[0][0] == "DAG_INPUT":
                    inputs[edge[1][1]] = base_inputs[edge[0][1]]
                else:
                    if list(outputs[edge[0][0]].keys()) == [""]:
                        outputs[edge[0][0]][edge[0][1]] = outputs[edge[0][0]][""]
                    inputs[edge[1][1]] = outputs[edge[0][0]][edge[0][1]]

            assert set(inputs_needed).issubset(
                set(inputs.keys())
            ), f"Inputs needed: {inputs_needed}, Inputs given: {inputs.keys()}"
            inputs_normalized = {k: str(v) for k, v in inputs.items()}
            if node_name not in frozen_node_names:
                outputs_unstructured = await self.nodes[node_name].arun(
                    inputs=inputs_normalized, verbose=verbose
                )
                outputs[node_name] = {}
                if isinstance(outputs_unstructured, str):
                    outputs[node_name][""] = outputs_unstructured
                else:
                    dictified = (
                        outputs_unstructured.dict()
                        if isinstance(outputs_unstructured, BaseModel)
                        else outputs_unstructured
                    )
                    for key in dictified.keys():
                        outputs[node_name][key] = dictified[key]
                stage_records.append(await self.nodes[node_name].get_record())
            else:
                for record in dag_record.stage_records:
                    if (
                        record.name == node_name
                    ):  # TODO: this should be a runnable record
                        if isinstance(record.outputs, str):
                            outputs[node_name] = {}
                            outputs[node_name][""] = record.outputs
                        else:
                            outputs[node_name] = record.outputs
                        stage_records.append(record)
                        break

        name_graph = self.compile()
        dag_record = DagRecord(stage_records=stage_records, name_graph=name_graph)

        dag_outputs = {}
        for edge in self.edges:
            if edge[1][0] == "DAG_OUTPUT":
                if list(outputs[edge[0][0]].keys()) == [""]:
                    outputs[edge[0][0]][edge[0][1]] = outputs[edge[0][0]][""]
                dag_outputs[edge[1][1]] = outputs[edge[0][0]][edge[0][1]]

        for k in list(
            dag_outputs.keys()
        ):  # Use list to create a copy of keys for safe iteration
            if k in self.output_aliases:
                dag_outputs[self.output_aliases[k]] = dag_outputs[k]
                del dag_outputs[k]
        if verbose:
            return dag_outputs, dag_record

        return dag_outputs

    async def arun_standard(self, base_inputs: Dict[str, Any], verbose: bool = False):
        unaliased_base_inputs = copy.deepcopy(base_inputs)
        for k, v in unaliased_base_inputs.items():
            if k in self.input_aliases:
                del base_inputs[k]
                base_inputs[self.input_aliases[k]] = v

        assert set(
            [edge[0][1] for edge in self.edges if edge[0][0] == "DAG_INPUT"]
        ).issubset(
            set(base_inputs.keys())
        ), f"Inputs needed: {[edge[0][1] for edge in self.edges if edge[0][0] == 'DAG_INPUT']}, Inputs given: {base_inputs.keys()}"
        stage_records = []
        g = self.compile()
        outputs = {}
        for node_name in nx.topological_sort(g):
            if node_name in ["DAG_INPUT", "DAG_OUTPUT"]:
                continue
            node = self.nodes[node_name]
            inputs_needed = node.get_inputs()
            inputs = {}
            for edge in self.edges:
                if edge[1][0] != node_name:
                    continue
                if edge[1][0] == "DAG_OUTPUT":
                    continue
                if edge[0][0] == "DAG_INPUT":
                    inputs[edge[1][1]] = base_inputs[edge[0][1]]
                else:
                    if list(outputs[edge[0][0]].keys()) == [""]:
                        # print("Warning - fix this")
                        outputs[edge[0][0]][edge[0][1]] = outputs[edge[0][0]][""]
                    inputs[edge[1][1]] = outputs[edge[0][0]][edge[0][1]]
            assert set(inputs_needed).issubset(
                set(inputs.keys())
            ), f"Inputs needed: {inputs_needed}, Inputs given: {inputs.keys()}"
            inputs_normalized = {k: str(v) for k, v in inputs.items()}
            outputs_unstructured = await self.nodes[node_name].arun(
                inputs=inputs_normalized, verbose=verbose
            )
            assert isinstance(outputs_unstructured, str) or isinstance(
                outputs_unstructured, BaseModel
            ), f"Outputs unstructured: {outputs_unstructured}"
            outputs[node_name] = {}
            if isinstance(outputs_unstructured, str):
                outputs[node_name][""] = outputs_unstructured
            else:
                dictified = (
                    outputs_unstructured.dict()
                    if isinstance(outputs_unstructured, BaseModel)
                    else outputs_unstructured
                )
                for key in dictified.keys():
                    outputs[node_name][key] = dictified[key]

            stage_records.append(await self.nodes[node_name].get_record())
        name_graph = self.compile()
        dag_record = DagRecord(stage_records=stage_records, name_graph=name_graph)

        dag_outputs = {}
        for edge in self.edges:
            if edge[1][0] == "DAG_OUTPUT":
                if list(outputs[edge[0][0]].keys()) in [[""], [""]]:
                    outputs[edge[0][0]][edge[0][1]] = outputs[edge[0][0]][""]
                elif (
                    not edge[0][1] in list(outputs[edge[0][0]].keys())
                    and edge[0][1] == "answer"
                ):
                    outputs[edge[0][0]]["answer"] = outputs[edge[0][0]]
                dag_outputs[edge[1][1]] = outputs[edge[0][0]][edge[0][1]]

        for k in list(dag_outputs.keys()):
            if k in self.output_aliases:
                dag_outputs[self.output_aliases[k]] = dag_outputs[k]
                del dag_outputs[k]

        if verbose:
            return dag_outputs, dag_record
        return dag_outputs

    def run_standard(self, base_inputs: Dict[str, Any], verbose: bool = False):
        unaliased_base_inputs = copy.deepcopy(base_inputs)
        for k, v in unaliased_base_inputs.items():
            if k in self.input_aliases:
                del base_inputs[k]
                base_inputs[self.input_aliases[k]] = v

        assert set(
            [edge[0][1] for edge in self.edges if edge[0][0] == "DAG_INPUT"]
        ).issubset(
            set(base_inputs.keys())
        ), f"Inputs needed: {[edge[0][1] for edge in self.edges if edge[0][0] == 'DAG_INPUT']}, Inputs given: {base_inputs.keys()}"
        stage_records = []
        g = self.compile()
        outputs = {}
        for node_name in nx.topological_sort(g):
            if node_name in ["DAG_INPUT", "DAG_OUTPUT"]:
                continue
            node = self.nodes[node_name]
            inputs_needed = node.get_inputs()  # Assuming get_inputs can be synchronous
            inputs = {}
            for edge in self.edges:
                if edge[1][0] != node_name:
                    continue
                if edge[1][0] == "DAG_OUTPUT":
                    continue
                if edge[0][0] == "DAG_INPUT":
                    inputs[edge[1][1]] = base_inputs[edge[0][1]]
                else:
                    if list(outputs[edge[0][0]].keys()) == [""]:
                        outputs[edge[0][0]][edge[0][1]] = outputs[edge[0][0]][""]
                    inputs[edge[1][1]] = outputs[edge[0][0]][edge[0][1]]
            assert set(inputs_needed).issubset(
                set(inputs.keys())
            ), f"Inputs needed: {inputs_needed}, Inputs given: {inputs.keys()}"
            inputs_normalized = {k: str(v) for k, v in inputs.items()}
            outputs_unstructured = node.run(
                inputs=inputs_normalized, verbose=verbose
            )  # Assuming run can be synchronous
            outputs[node_name] = {}

            if isinstance(outputs_unstructured, str):
                outputs[node_name][""] = outputs_unstructured
            else:
                dictified = (
                    outputs_unstructured.dict()
                    if isinstance(outputs_unstructured, BaseModel)
                    else outputs_unstructured
                )
                for key in dictified.keys():
                    outputs[node_name][key] = dictified[key]

            stage_records.append(node.get_record_sync())
        name_graph = self.compile()
        dag_record = DagRecord(stage_records=stage_records, name_graph=name_graph)

        dag_outputs = {}
        for edge in self.edges:
            if edge[1][0] == "DAG_OUTPUT":
                if list(outputs[edge[0][0]].keys()) in [[""], [""]]:
                    outputs[edge[0][0]][edge[0][1]] = outputs[edge[0][0]][""]
                elif (
                    not edge[0][1] in list(outputs[edge[0][0]].keys())
                    and edge[0][1] == "answer"
                ):
                    outputs[edge[0][0]]["answer"] = outputs[edge[0][0]]
                dag_outputs[edge[1][1]] = outputs[edge[0][0]][edge[0][1]]

        for k in list(
            dag_outputs.keys()
        ):  # Use list to create a copy of keys for safe iteration
            if k in self.output_aliases:
                dag_outputs[self.output_aliases[k]] = dag_outputs[k]
                del dag_outputs[k]

        if verbose:
            return dag_outputs, dag_record
        return dag_outputs



import random
import json
from collections import namedtuple

import messenger.envs.config as config

Descr = namedtuple("Description", ["entity", "role", "type"])


class TextManual:
    """
    Class which implements methods that allow environments to construct
    text manuals for games in Messenger.
    """

    def __init__(self, json_path):
        with open(json_path, "r") as f:
            self.descriptors = json.load(f)

    def get_descriptor(self, entity, role, entity_type, no_type_p=0.15):
        """
        Get a descriptor using the templates.
        Parameters:
        entity: The object that is being described (e.g. alien)
        role: The role the the object plays in the env. (e.g. enemy)
        entity_type:
            The object type (e.g. chaser)
        no_type_p:
            The probability of returning a descriptor that does not have
            any type information (only if entity_type is not None).
        """
        if random.random() < no_type_p:  # no type information
            return random.choice(self.descriptors[entity][role]["unknown"])

        else:
            return random.choice(self.descriptors[entity][role][entity_type])

    def get_document(
        self,
        enemy,
        message,
        goal,
        shuffle=False,
        enemy_type=None,
        message_type=None,
        goal_type=None,
        append=False,
        delete=False,
        **kwargs,
    ):
        """
        Makes a document for Messenger using the specified parameters.
        If no type is provided, a random type will be selected.

        Parameters:
        append:
            If True, append an extraneous sentence to the document describing a
            random object that is not in {enemy, message, goal}.
        delete: If True, Delete a random descriptor from the document.
        shuffle:
            If True, shuffles the order of the descriptors
        kwargs:
            All other kwargs go to get_descriptor()
        """

        document = [
            self.get_descriptor(
                entity=enemy, entity_type=enemy_type, role="enemy", **kwargs
            ),
            self.get_descriptor(
                entity=message, entity_type=message_type, role="message", **kwargs
            ),
            self.get_descriptor(
                entity=goal, entity_type=goal_type, role="goal", **kwargs
            ),
        ]

        if delete:  # delete a random descriptor
            document = random.sample(document, 2)

        if append:
            # choose an object not in {enemy, message, goal}
            valid_objs = [
                obj.name
                for obj in config.NPCS
                if obj.name not in [enemy, message, goal]
            ]
            rand_obj = random.choice(valid_objs)
            result = None
            while result is None:
                try:
                    result = self.get_descriptor(
                        entity=rand_obj,
                        role=random.choice(("enemy", "message", "goal")),
                        entity_type=random.choice(("chaser", "fleeing", "immovable")),
                        **kwargs,
                    )
                except:
                    pass
            document.append(result)

        if shuffle:
            document = random.sample(document, len(document))

        return document

    def get_document_plus(self, *args, **kwargs):
        """
        Makes a document for Messenger using the specified parameters.

        Parameters:
        args: List of Descrip namedtuples
        """

        document = []

        for descrip in args:
            document.append(
                self.get_descriptor(
                    entity=descrip.entity,
                    role=descrip.role,
                    entity_type=descrip.type,
                    no_type_p=0,
                    **kwargs,
                )
            )

        return document

    def get_decoy_descriptor(self, entity, not_of_role, not_of_type, **kwargs):
        """
        Get a description about the entity where the entity is not of role not_of_role
        and not of type not_of_type
        """
        possible_roles = [x for x in ("message", "goal", "enemy") if x != not_of_role]
        random.shuffle(possible_roles)
        selected_type = random.choice(
            [x for x in ("chaser", "fleeing", "immovable") if x != not_of_type]
        )

        for role in possible_roles:
            try:
                return self.get_descriptor(
                    entity=entity,
                    role=role,
                    entity_type=selected_type,
                    no_type_p=0,
                    **kwargs,
                )
            except:
                continue
        raise Exception("decoy description with impossible constraints")


if __name__ == "__main__":
    # just some quick and dirty tests
    from pathlib import Path

    text_json = Path(__file__).parent.joinpath("texts", "text_train.json")
    manual = TextManual(json_path=text_json)
    descriptions = (
        Descr(entity="airplane", role="goal", type="chaser"),
        Descr(entity="airplane", role="message", type="fleeing"),
        Descr(entity="dog", role="enemy", type="immovable"),
        Descr(entity="dog", role="goal", type="fleeing"),
        Descr(entity="mage", role="goal", type="chaser"),
        Descr(entity="mage", role="message", type="immovable"),
    )
    print(manual.get_document_plus(*descriptions))



import functools

import numpy as np
import opensimplex

from . import constants
from . import objects


def generate_world(world, player):
    simplex = opensimplex.OpenSimplex(seed=world.random.randint(0, 2**31 - 1))
    tunnels = np.zeros(world.area, bool)
    for x in range(world.area[0]):
        for y in range(world.area[1]):
            _set_material(world, (x, y), player, tunnels, simplex)
    for x in range(world.area[0]):
        for y in range(world.area[1]):
            _set_object(world, (x, y), player, tunnels)


def _set_material(world, pos, player, tunnels, simplex):
    x, y = pos
    simplex = functools.partial(_simplex, simplex)
    uniform = world.random.uniform
    start = 4 - np.sqrt((x - player.pos[0]) ** 2 + (y - player.pos[1]) ** 2)
    start += 2 * simplex(x, y, 8, 3)
    start = 1 / (1 + np.exp(-start))
    water = simplex(x, y, 3, {15: 1, 5: 0.15}, False) + 0.1
    water -= 2 * start
    mountain = simplex(x, y, 0, {15: 1, 5: 0.3})
    mountain -= 4 * start + 0.3 * water
    if start > 0.5:
        world[x, y] = "grass"
    elif mountain > 0.15:
        if simplex(x, y, 6, 7) > 0.15 and mountain > 0.3:  # cave
            world[x, y] = "path"
        elif simplex(2 * x, y / 5, 7, 3) > 0.4:  # horizonal tunnle
            world[x, y] = "path"
            tunnels[x, y] = True
        elif simplex(x / 5, 2 * y, 7, 3) > 0.4:  # vertical tunnle
            world[x, y] = "path"
            tunnels[x, y] = True
        elif simplex(x, y, 1, 8) > 0 and uniform() > 0.85:
            world[x, y] = "coal"
        elif simplex(x, y, 2, 6) > 0.4 and uniform() > 0.75:
            world[x, y] = "iron"
        elif mountain > 0.18 and uniform() > 0.994:
            world[x, y] = "diamond"
        elif mountain > 0.3 and simplex(x, y, 6, 5) > 0.35:
            world[x, y] = "lava"
        else:
            world[x, y] = "stone"
    elif 0.25 < water <= 0.35 and simplex(x, y, 4, 9) > -0.2:
        world[x, y] = "sand"
    elif 0.3 < water:
        world[x, y] = "water"
    else:  # grassland
        if simplex(x, y, 5, 7) > 0 and uniform() > 0.8:
            world[x, y] = "tree"
        else:
            world[x, y] = "grass"


def _set_object(world, pos, player, tunnels):
    x, y = pos
    uniform = world.random.uniform
    dist = np.sqrt((x - player.pos[0]) ** 2 + (y - player.pos[1]) ** 2)
    material, _ = world[x, y]
    if material not in constants.walkable:
        pass
    elif dist > 3 and material == "grass" and uniform() > 0.985:
        world.add(objects.Cow(world, (x, y)))
    elif dist > 10 and uniform() > 0.993:
        world.add(objects.Zombie(world, (x, y), player))
    elif material == "path" and tunnels[x, y] and uniform() > 0.95:
        world.add(objects.Skeleton(world, (x, y), player))


def _simplex(simplex, x, y, z, sizes, normalize=True):
    if not isinstance(sizes, dict):
        sizes = {sizes: 1}
    value = 0
    for size, weight in sizes.items():
        if hasattr(simplex, "noise3d"):
            noise = simplex.noise3d(x / size, y / size, z)
        else:
            noise = simplex.noise3(x / size, y / size, z)
        value += weight * noise
    if normalize:
        value /= sum(sizes.values())
    return value



import glob
import os

from apropos.src.core.lms.api_caching_fxns import SafeCache

cache = SafeCache("apropos/src/lms/.cache", "apropos/src/lms/.cache.db")


def print_cache_folder_sizes():
    cache_folders = glob.glob("apropos/src/lms/.cache")
    db_files = glob.glob("apropos/src/lms/.*.db")

    for db_file in db_files:
        if os.path.exists(db_file):
            db_size = os.path.getsize(db_file)
            print(f"{db_file} (SQLite): {db_size / (1024 * 1024):.2f} MB")

    for folder in cache_folders:
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(folder):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)
        print(f"{folder}: {total_size / (1024 * 1024):.2f} MB")


# print_cache_folder_sizes()



import numpy as np

from . import constants
from . import engine


class Object:
    def __init__(self, world, pos):
        self.world = world
        self.pos = np.array(pos)
        self.random = world.random
        self.inventory = {"health": 0}
        self.removed = False

    @property
    def texture(self):
        raise "unknown"

    @property
    def walkable(self):
        return constants.walkable

    @property
    def health(self):
        return self.inventory["health"]

    @health.setter
    def health(self, value):
        self.inventory["health"] = max(0, value)

    @property
    def all_dirs(self):
        return ((-1, 0), (+1, 0), (0, -1), (0, +1))

    def move(self, direction):
        direction = np.array(direction)
        target = self.pos + direction
        if self.is_free(target):
            self.world.move(self, target)
            return True
        return False

    def is_free(self, target, materials=None):
        materials = self.walkable if materials is None else materials
        material, obj = self.world[target]
        return obj is None and material in materials

    def distance(self, target):
        if hasattr(target, "pos"):
            target = target.pos
        return np.abs(target - self.pos).sum()

    def toward(self, target, long_axis=True):
        if hasattr(target, "pos"):
            target = target.pos
        offset = target - self.pos
        dists = np.abs(offset)
        if dists[0] > dists[1] if long_axis else dists[0] <= dists[1]:
            return np.array((np.sign(offset[0]), 0))
        else:
            return np.array((0, np.sign(offset[1])))

    def random_dir(self):
        return self.all_dirs[self.random.randint(0, 4)]


class Player(Object):
    def __init__(self, world, pos):
        super().__init__(world, pos)
        self.facing = (0, 1)
        self.inventory = {
            name: info["initial"] for name, info in constants.items.items()
        }
        self.achievements = {name: 0 for name in constants.achievements}
        self.action = "noop"
        self.sleeping = False
        self._last_health = self.health
        self._hunger = 0
        self._thirst = 0
        self._fatigue = 0
        self._recover = 0

    @property
    def texture(self):
        if self.sleeping:
            return "player-sleep"
        return {
            (-1, 0): "player-left",
            (+1, 0): "player-right",
            (0, -1): "player-up",
            (0, +1): "player-down",
        }[tuple(self.facing)]

    @property
    def walkable(self):
        return constants.walkable + ["lava"]

    def update(self):
        target = (self.pos[0] + self.facing[0], self.pos[1] + self.facing[1])
        material, obj = self.world[target]
        action = self.action
        if self.sleeping:
            if self.inventory["energy"] < constants.items["energy"]["max"]:
                action = "sleep"
            else:
                self.sleeping = False
                self.achievements["wake_up"] += 1
        if action == "noop":
            pass
        elif action.startswith("move_"):
            self._move(action[len("move_") :])
        elif action == "do" and obj:
            self._do_object(obj)
        elif action == "do":
            self._do_material(target, material)
        elif action == "sleep":
            if self.inventory["energy"] < constants.items["energy"]["max"]:
                self.sleeping = True
        elif action.startswith("place_"):
            self._place(action[len("place_") :], target, material)
        elif action.startswith("make_"):
            self._make(action[len("make_") :])
        self._update_life_stats()
        self._degen_or_regen_health()
        for name, amount in self.inventory.items():
            maxmium = constants.items[name]["max"]
            self.inventory[name] = max(0, min(amount, maxmium))
        # This needs to happen after the inventory states are clamped
        # because it involves the health water inventory count.
        self._wake_up_when_hurt()

    def _update_life_stats(self):
        self._hunger += 0.5 if self.sleeping else 1
        if self._hunger > 25:
            self._hunger = 0
            self.inventory["food"] -= 1
        self._thirst += 0.5 if self.sleeping else 1
        if self._thirst > 20:
            self._thirst = 0
            self.inventory["drink"] -= 1
        if self.sleeping:
            self._fatigue = min(self._fatigue - 1, 0)
        else:
            self._fatigue += 1
        if self._fatigue < -10:
            self._fatigue = 0
            self.inventory["energy"] += 1
        if self._fatigue > 30:
            self._fatigue = 0
            self.inventory["energy"] -= 1

    def _degen_or_regen_health(self):
        necessities = (
            self.inventory["food"] > 0,
            self.inventory["drink"] > 0,
            self.inventory["energy"] > 0 or self.sleeping,
        )
        if all(necessities):
            self._recover += 2 if self.sleeping else 1
        else:
            self._recover -= 0.5 if self.sleeping else 1
        if self._recover > 25:
            self._recover = 0
            self.health += 1
        if self._recover < -15:
            self._recover = 0
            self.health -= 1

    def _wake_up_when_hurt(self):
        if self.health < self._last_health:
            self.sleeping = False
        self._last_health = self.health

    def _move(self, direction):
        directions = dict(left=(-1, 0), right=(+1, 0), up=(0, -1), down=(0, +1))
        self.facing = directions[direction]
        self.move(self.facing)
        if self.world[self.pos][0] == "lava":
            self.health = 0

    def _do_object(self, obj):
        damage = max(
            [
                1,
                self.inventory["wood_sword"] and 2,
                self.inventory["stone_sword"] and 3,
                self.inventory["iron_sword"] and 5,
            ]
        )
        if isinstance(obj, Plant):
            if obj.ripe:
                obj.grown = 0
                self.inventory["food"] += 4
                self.achievements["eat_plant"] += 1
        if isinstance(obj, Fence):
            self.world.remove(obj)
            self.inventory["fence"] += 1
            self.achievements["collect_fence"] += 1
        if isinstance(obj, Zombie):
            obj.health -= damage
            if obj.health <= 0:
                self.achievements["defeat_zombie"] += 1
        if isinstance(obj, Skeleton):
            obj.health -= damage
            if obj.health <= 0:
                self.achievements["defeat_skeleton"] += 1
        if isinstance(obj, Cow):
            obj.health -= damage
            if obj.health <= 0:
                self.inventory["food"] += 6
                self.achievements["eat_cow"] += 1
                # TODO: Keep track of previous inventory state to do this in a more
                # general way.
                self._hunger = 0

    def _do_material(self, target, material):
        if material == "water":
            # TODO: Keep track of previous inventory state to do this in a more
            # general way.
            self._thirst = 0
        info = constants.collect.get(material)
        if not info:
            return
        for name, amount in info["require"].items():
            if self.inventory[name] < amount:
                return
        self.world[target] = info["leaves"]
        if self.random.uniform() <= info.get("probability", 1):
            for name, amount in info["receive"].items():
                self.inventory[name] += amount
                self.achievements[f"collect_{name}"] += 1

    def _place(self, name, target, material):
        if self.world[target][1]:
            return
        info = constants.place[name]
        if material not in info["where"]:
            return
        if any(self.inventory[k] < v for k, v in info["uses"].items()):
            return
        for item, amount in info["uses"].items():
            self.inventory[item] -= amount
        if info["type"] == "material":
            self.world[target] = name
        elif info["type"] == "object":
            cls = {
                "fence": Fence,
                "plant": Plant,
            }[name]
            self.world.add(cls(self.world, target))
        self.achievements[f"place_{name}"] += 1

    def _make(self, name):
        nearby, _ = self.world.nearby(self.pos, 1)
        info = constants.make[name]
        if not all(util in nearby for util in info["nearby"]):
            return
        if any(self.inventory[k] < v for k, v in info["uses"].items()):
            return
        for item, amount in info["uses"].items():
            self.inventory[item] -= amount
        self.inventory[name] += info["gives"]
        self.achievements[f"make_{name}"] += 1


class Cow(Object):
    def __init__(self, world, pos):
        super().__init__(world, pos)
        self.health = 3

    @property
    def texture(self):
        return "cow"

    def update(self):
        if self.health <= 0:
            self.world.remove(self)
        if self.random.uniform() < 0.5:
            direction = self.random_dir()
            self.move(direction)


class Zombie(Object):
    def __init__(self, world, pos, player):
        super().__init__(world, pos)
        self.player = player
        self.health = 5
        self.cooldown = 0

    @property
    def texture(self):
        return "zombie"

    def update(self):
        if self.health <= 0:
            self.world.remove(self)
        dist = self.distance(self.player)
        if dist <= 8 and self.random.uniform() < 0.9:
            self.move(self.toward(self.player, self.random.uniform() < 0.8))
        else:
            self.move(self.random_dir())
        dist = self.distance(self.player)
        if dist <= 1:
            if self.cooldown:
                self.cooldown -= 1
            else:
                if self.player.sleeping:
                    damage = 7
                else:
                    damage = 2
                self.player.health -= damage
                self.cooldown = 5


class Skeleton(Object):
    def __init__(self, world, pos, player):
        super().__init__(world, pos)
        self.player = player
        self.health = 3
        self.reload = 0

    @property
    def texture(self):
        return "skeleton"

    def update(self):
        if self.health <= 0:
            self.world.remove(self)
        self.reload = max(0, self.reload - 1)
        dist = self.distance(self.player.pos)
        if dist <= 3:
            moved = self.move(-self.toward(self.player, self.random.uniform() < 0.6))
            if moved:
                return
        if dist <= 5 and self.random.uniform() < 0.5:
            self._shoot(self.toward(self.player))
        elif dist <= 8 and self.random.uniform() < 0.3:
            self.move(self.toward(self.player, self.random.uniform() < 0.6))
        elif self.random.uniform() < 0.2:
            self.move(self.random_dir())

    def _shoot(self, direction):
        if self.reload > 0:
            return
        if direction[0] == 0 and direction[1] == 0:
            return
        pos = self.pos + direction
        if self.is_free(pos, Arrow.walkable):
            self.world.add(Arrow(self.world, pos, direction))
            self.reload = 4


class Arrow(Object):
    def __init__(self, world, pos, facing):
        super().__init__(world, pos)
        self.facing = facing

    @property
    def texture(self):
        return {
            (-1, 0): "arrow-left",
            (+1, 0): "arrow-right",
            (0, -1): "arrow-up",
            (0, +1): "arrow-down",
        }[tuple(self.facing)]

    @engine.staticproperty
    def walkable():
        return constants.walkable + ["water", "lava"]

    def update(self):
        target = self.pos + self.facing
        material, obj = self.world[target]
        if obj:
            obj.health -= 2
            self.world.remove(self)
        elif material not in self.walkable:
            self.world.remove(self)
            if material in ["table", "furnace"]:
                self.world[target] = "path"
        else:
            self.move(self.facing)


class Plant(Object):
    def __init__(self, world, pos):
        super().__init__(world, pos)
        self.health = 1
        self.grown = 0

    @property
    def texture(self):
        if self.ripe:
            return "plant-ripe"
        else:
            return "plant"

    @property
    def ripe(self):
        return self.grown > 300

    def update(self):
        self.grown += 1
        objs = [self.world[self.pos + dir_][1] for dir_ in self.all_dirs]
        if any(isinstance(obj, (Zombie, Skeleton, Cow)) for obj in objs):
            self.health -= 1
        if self.health <= 0:
            self.world.remove(self)


class Fence(Object):
    def __init__(self, world, pos):
        super().__init__(world, pos)

    @property
    def texture(self):
        return "fence"

    def update(self):
        pass



"""
Implements wrappers on top of the basic messenger environments
"""

import random

from messenger.envs.base import MessengerEnv
from messenger.envs.stage_one import StageOne
from messenger.envs.stage_two import StageTwo
from messenger.envs.stage_three import StageThree


class TwoEnvWrapper(MessengerEnv):
    """
    Switches between two Messenger environments
    """

    def __init__(
        self, stage: int, split_1: str, split_2: str, prob_env_1=0.5, **kwargs
    ):
        super().__init__()
        if stage == 1:
            self.env_1 = StageOne(split=split_1, **kwargs)
            self.env_2 = StageOne(split=split_2, **kwargs)
        elif stage == 2:
            self.env_1 = StageTwo(split=split_1, **kwargs)
            self.env_2 = StageTwo(split=split_2, **kwargs)
        elif stage == 3:
            self.env_1 = StageThree(split=split_1, **kwargs)
            self.env_2 = StageThree(split=split_2, **kwargs)

        self.prob_env_1 = prob_env_1
        self.cur_env = None

    def reset(self, **kwargs):
        if random.random() < self.prob_env_1:
            self.cur_env = self.env_1
        else:
            self.cur_env = self.env_2
        return self.cur_env.reset(**kwargs)

    def step(self, action):
        return self.cur_env.step(action)



import asyncio
import logging
import os
from typing import Dict, List

import backoff
import google.generativeai as genai

from apropos.src.core.lms.cache_init import cache
from apropos.src.core.lms.vendors.base import BaseProvider
from apropos.src.core.lms.vendors.json_structured_outputs.core import (
    add_json_instructions_to_messages,
    extract_pydantic_model_from_response_sync,
    extract_pydantic_model_from_response_async,
)
import google.api_core.exceptions

# Suppress all logging from google.generativeai
logging.getLogger("google.generativeai").setLevel(logging.ERROR)
os.environ["GRPC_VERBOSITY"] = "ERROR"
# Suppress TensorFlow logging
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

# Suppress other warnings
import warnings

warnings.filterwarnings("ignore")

BACKOFF_TOLERANCE = 10


class DeepmindAPIProvider(BaseProvider):
    def __init__(self):
        self.supports_response_model = True

    @backoff.on_exception(
        backoff.expo,
        (
            Exception,
            google.api_core.exceptions.ResourceExhausted,
        ),  # Replace with specific exceptions if known
        max_tries=BACKOFF_TOLERANCE,
    )
    async def hit_gemini_async(
        self,
        messages: List[Dict],
        temperature: float = 0,
        model_name: str = "gemini-1.5-flash",
    ) -> str:
        code_generation_model = genai.GenerativeModel(
            model_name=model_name,
            generation_config={"temperature": temperature},
            system_instruction=messages[0]["content"],
        )
        result = await code_generation_model.generate_content_async(
            messages[1]["content"]
        )
        try:
            return result.text
        except Exception as e:
            print("Gemini failed", e)
            print(result)
            return "Gemini failed"

    @backoff.on_exception(
        backoff.expo,
        (Exception),  # Replace with specific exceptions if known
        max_tries=BACKOFF_TOLERANCE,
    )
    async def async_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        messages_with_json_formatting_instructions = add_json_instructions_to_messages(
            messages, response_model
        )
        hit = cache.hit_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
        )
        if hit:
            return hit
        raw_text_api_response = await self.async_chat_completion(
            messages_with_json_formatting_instructions, model, temperature, max_tokens
        )
        structured_api_response = await extract_pydantic_model_from_response_async(
            raw_text_api_response, response_model
        )
        cache.add_to_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
            structured_api_response,
        )
        return structured_api_response

    @backoff.on_exception(
        backoff.expo,
        (
            Exception,
            google.api_core.exceptions.ResourceExhausted,
        ),  # Replace with specific exceptions if known
        max_tries=BACKOFF_TOLERANCE,
    )
    def sync_chat_completion_with_response_model(
        self, messages, model, temperature, max_tokens, response_model
    ):
        messages_with_json_formatting_instructions = add_json_instructions_to_messages(
            messages, response_model
        )
        hit = cache.hit_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
        )
        if hit:
            return hit
        raw_text_api_response = self.sync_chat_completion(
            messages_with_json_formatting_instructions, model, temperature, max_tokens
        )
        structured_api_response = extract_pydantic_model_from_response_sync(
            raw_text_api_response, response_model
        )
        cache.add_to_cache(
            messages_with_json_formatting_instructions,
            model,
            temperature,
            response_model,
            structured_api_response,
        )
        return structured_api_response

    @backoff.on_exception(
        backoff.expo,
        (Exception),  # Replace with specific exceptions if known
        max_tries=BACKOFF_TOLERANCE,
    )
    async def async_chat_completion(self, messages, model, temperature, max_tokens):
        hit = cache.hit_cache(messages, model, temperature, None)
        if hit:
            return hit
        result = await self.hit_gemini_async(
            messages=messages,
            model_name=model,
            temperature=temperature,
        )
        cache.add_to_cache(messages, model, temperature, None, result)
        return result

    @backoff.on_exception(
        backoff.expo,
        (Exception),  # Replace with specific exceptions if known
        max_tries=BACKOFF_TOLERANCE,
    )
    def sync_chat_completion(self, messages, model, temperature, max_tokens):
        hit = cache.hit_cache(messages, model, temperature, None)
        if hit:
            return hit
        result = asyncio.run(
            self.hit_gemini_async(
                messages=messages,
                model_name=model,
                temperature=temperature,
            )
        )
        cache.add_to_cache(messages, model, temperature, None, result)
        return result


if __name__ == "__main__":
    import asyncio

    messages = [
        {
            "role": "system",
            "content": "You are a helpful assistant that can answer questions about the capital of France.",
        },
        {"role": "user", "content": " What is t he capital of France? "},
    ]
    response = asyncio.run(
        DeepmindAPIProvider().async_chat_completion(
            messages=messages,
            model="gemini-1.5-flash",
            temperature=0.0,
            max_tokens=150,
        )
    )
    print(response)



import argparse
import pathlib
import time

import numpy as np

import crafter


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--seed", type=int, default=None)
    parser.add_argument("--area", nargs=2, type=int, default=(64, 64))
    parser.add_argument("--length", type=int, default=10000)
    parser.add_argument("--health", type=int, default=9)
    parser.add_argument("--record", type=pathlib.Path, default=None)
    parser.add_argument("--episodes", type=int, default=1)
    args = parser.parse_args()

    random = np.random.RandomState(args.seed)
    crafter.constants.items["health"]["max"] = args.health
    crafter.constants.items["health"]["initial"] = args.health
    env = crafter.Env(area=args.area, length=args.length, seed=args.seed)
    env = crafter.Recorder(env, args.record)

    for _ in range(args.episodes):
        start = time.time()
        obs = env.reset()
        print("")
        print(f"Reset time: {1000*(time.time()-start):.2f}ms")
        print("Coal exist:    ", env._world.count("coal"))
        print("Iron exist:    ", env._world.count("iron"))
        print("Diamonds exist:", env._world.count("diamond"))

        start = time.time()
        done = False
        while not done:
            action = random.randint(0, env.action_space.n)
            obs, reward, done, info = env.step(action)
        duration = time.time() - start
        step = env._step
        print(f"Step time: {1000*duration/step:.2f}ms ({int(step/duration)} FPS)")
        print("Episode length:", step)


if __name__ == "__main__":
    main()



import copy
from typing import Callable, Dict, List, Optional, Type

from apropos.src.bench.base import Benchmark
from apropos.src.core.programs.dag import LM_DAG
from apropos.src.core.programs.prompt import Demonstration


class DAGOptimizer:
    student_program: LM_DAG
    teacher_program: Optional[LM_DAG]
    dataset_handler: Type[Benchmark]

    def get_fewshot_program(
        self, demonstrations_by_node_name: Dict[str, List[Demonstration]]
    ):
        fewshot_program = copy.deepcopy(self.student_program)
        for node_name, demonstrations_for_stage in demonstrations_by_node_name.items():
            optimal_demos = demonstrations_for_stage
            node = fewshot_program.nodes[node_name]
            node.get_input_fields.prompt.demonstrations = optimal_demos
            node.transform.prompt.demonstrations = optimal_demos
            node.produce_stage_record.prompt.demonstrations = optimal_demos
        return fewshot_program

    # Annotator must send DAG to annotated DAG
    async def bootstrap_demonstrations(
        self, annotator: Optional[Callable] = None, n: int = 20, patches=["A", "B"]
    ) -> Dict[str, List[Demonstration]]:
        if self.teacher_program:
            candidate = copy.deepcopy(self.teacher_program)
        else:
            candidate = copy.deepcopy(self.student_program)
        scores, dag_records = await self.dataset_handler.score_dag(
            candidate, n=n, verbose=True, patches=patches, split="train"
        )
        fewshot_demonstrations_by_node = {
            node_key: [] for node_key in candidate.nodes.keys()
        }
        for dag_record, score in zip(dag_records, scores):
            if score == True:
                if not annotator:
                    annotated_stage_demonstrations = (
                        dag_record.trivially_annotate().to_stage_demonstrations()
                    )
                else:
                    annotated_stage_demonstrations = annotator(
                        dag_record
                    ).to_stage_demonstrations()
                for i, stage_demonstrations in enumerate(
                    annotated_stage_demonstrations
                ):
                    fewshot_demonstrations_by_node[
                        list(candidate.nodes.keys())[i]
                    ].append(stage_demonstrations)
        return fewshot_demonstrations_by_node



craftax_classic_game_rules = {
    "World Information": {
        "Actions with Requirements": {
            "Movement": {
                "left": {"Requirements": "Flat ground west of the agent."},
                "right": {"Requirements": "Flat ground east of the agent."},
                "up": {"Requirements": "Flat ground north of the agent."},
                "down": {"Requirements": "Flat ground south of the agent."},
            },
            "Basic Actions": {
                "do": {
                    "Requirements": "Facing creature or material; have necessary tool."
                },
                "sleep": {"Requirements": "Energy level is below maximum."},
                "noop": {"Requirements": "Always applicable. (Means no operation.)"},
            },
            "Crafting and Building": {
                "place_stone": {"Requirements": "Stone in inventory."},
                "place_table": {"Requirements": "Wood in inventory."},
                "place_furnace": {"Requirements": "Stone in inventory."},
                "place_plant": {"Requirements": "Sapling in inventory."},
                "make_wood_pickaxe": {
                    "Requirements": "Nearby table; wood in inventory."
                },
                "make_stone_pickaxe": {
                    "Requirements": "Nearby table; wood, stone in inventory."
                },
                "make_iron_pickaxe": {
                    "Requirements": "Nearby table, furnace; wood, coal, iron in inventory."
                },
                "make_wood_sword": {"Requirements": "Nearby table; wood in inventory."},
                "make_stone_sword": {
                    "Requirements": "Nearby table; wood, stone in inventory."
                },
                "make_iron_sword": {
                    "Requirements": "Nearby table, furnace; wood, coal, iron in inventory."
                },
            },
        },
        "Hierarchy of Goals": {
            "Collect Wood": {
                "place_table": {
                    "make_wood_pickaxe": {
                        "Collect Stone": {
                            "place_stone": {},
                            "make_stone_pickaxe": {
                                "Collect Iron": {
                                    "make_iron_pickaxe": {
                                        "Collect Diamond": {},
                                        "place_furnace": {
                                            "make_iron_sword": {},
                                            "make_iron_pickaxe": {},
                                        },
                                    }
                                },
                                "Collect Coal": {},
                            },
                        }
                    },
                    "make_wood_sword": {},
                }
            },
            "Eat Cow": {},
            "Collect Sapling": {"place_plant": {"Eat Plant": {}}},
            "Collect Drink": {},
            "Defeat Zombie": {},
            "Defeat Skeleton": {},
            "Wake Up": {},
        },
    }
}
craftax_classic_action_dict = {
    "noop": 0,
    "left": 1,
    "right": 2,
    "up": 3,
    "down": 4,
    "do": 5,
    "sleep": 6,
    "place_stone": 7,
    "place_table": 8,
    "place_furnace": 9,
    "place_plant": 10,
    "make_wood_pickaxe": 11,
    "make_stone_pickaxe": 12,
    "make_iron_pickaxe": 13,
    "make_wood_sword": 14,
    "make_stone_sword": 15,
    "make_iron_sword": 16,
}

craftax_full_game_rules = {
    "World Information": {
        "Actions with Requirements": {
            "Movement": {
                "left": {"Requirements": "Flat ground west of the agent."},
                "right": {"Requirements": "Flat ground east of the agent."},
                "up": {"Requirements": "Flat ground north of the agent."},
                "down": {"Requirements": "Flat ground south of the agent."},
            },
            "Basic Actions": {
                "do": {
                    "Requirements": "Facing creature or material; have necessary tool."
                },
                "sleep": {"Requirements": "Energy level is below maximum."},
                "noop": {"Requirements": "Always applicable. (Means no operation.)"},
            },
            "Crafting and Building": {
                "place_stone": {
                    "Requirements": "Have stone in inventory; target location is empty or water."
                },
                "place_table": {
                    "Requirements": "Have 2 wood in inventory; target location is empty."
                },
                "place_furnace": {
                    "Requirements": "Have stone in inventory; target location is empty."
                },
                "place_torch": {
                    "Requirements": "Have torch in inventory; target location is valid for torch placement."
                },
                "place_plant": {
                    "Requirements": "Have sapling in inventory; target location is grass."
                },
                "make_wood_pickaxe": {
                    "Requirements": "At crafting table; have 1 wood in inventory."
                },
                "make_stone_pickaxe": {
                    "Requirements": "At crafting table; have 1 wood and 1 stone in inventory."
                },
                "make_iron_pickaxe": {
                    "Requirements": "At crafting table and furnace; have 1 wood, 1 stone, 1 iron, and 1 coal in inventory."
                },
                "make_diamond_pickaxe": {
                    "Requirements": "At crafting table; have 1 wood and 3 diamonds in inventory."
                },
                "make_wood_sword": {
                    "Requirements": "At crafting table; have 1 wood in inventory."
                },
                "make_stone_sword": {
                    "Requirements": "At crafting table; have 1 wood and 1 stone in inventory."
                },
                "make_iron_sword": {
                    "Requirements": "At crafting table and furnace; have 1 wood, 1 stone, 1 iron, and 1 coal in inventory."
                },
                "make_diamond_sword": {
                    "Requirements": "At crafting table; have 1 wood and 2 diamonds in inventory."
                },
                "make_iron_armour": {
                    "Requirements": "At crafting table and furnace; have 3 iron and 3 coal in inventory."
                },
                "make_diamond_armour": {
                    "Requirements": "At crafting table; have 3 diamonds in inventory."
                },
                "make_arrow": {
                    "Requirements": "At crafting table; have 1 wood and 1 stone in inventory."
                },
                "make_torch": {
                    "Requirements": "At crafting table; have 1 wood and 1 coal in inventory."
                },
            },
            "Combat and Tools": {
                "shoot_arrow": {"Requirements": "Have bow and arrows in inventory."},
                "cast_fireball": {
                    "Requirements": "Have learned fireball spell; have at least 2 mana."
                },
                "cast_iceball": {
                    "Requirements": "Have learned iceball spell; have at least 2 mana."
                },
                "enchant_bow": {
                    "Requirements": "At enchantment table; have bow, 1 gem (ruby/sapphire), and 9 mana."
                },
                "enchant_sword": {
                    "Requirements": "At enchantment table; have sword, 1 gem (ruby/sapphire), and 9 mana."
                },
                "enchant_armour": {
                    "Requirements": "At enchantment table; have armour, 1 gem (ruby/sapphire), and 9 mana."
                },
            },
            "Consumables": {
                "drink_potion_red": {"Requirements": "Have red potion in inventory."},
                "drink_potion_green": {
                    "Requirements": "Have green potion in inventory."
                },
                "drink_potion_blue": {"Requirements": "Have blue potion in inventory."},
                "drink_potion_pink": {"Requirements": "Have pink potion in inventory."},
                "drink_potion_cyan": {"Requirements": "Have cyan potion in inventory."},
                "drink_potion_yellow": {
                    "Requirements": "Have yellow potion in inventory."
                },
                "read_book": {"Requirements": "Have book in inventory."},
            },
            "Level Navigation": {
                "descend": {
                    "Requirements": "On down ladder; current level cleared of monsters."
                },
                "ascend": {"Requirements": "On up ladder."},
            },
            "Character Development": {
                "level_up_dexterity": {
                    "Requirements": "Have at least 1 XP; dexterity below maximum."
                },
                "level_up_strength": {
                    "Requirements": "Have at least 1 XP; strength below maximum."
                },
                "level_up_intelligence": {
                    "Requirements": "Have at least 1 XP; intelligence below maximum."
                },
            },
        }
    },
    "Hierarchy of Goals": {
        "Collect Wood": {
            "place_table": {
                "make_wood_pickaxe": {
                    "Collect Stone": {
                        "place_stone": {},
                        "make_stone_pickaxe": {
                            "Collect Iron": {
                                "make_iron_pickaxe": {
                                    "Collect Diamond": {
                                        "make_diamond_pickaxe": {},
                                        "make_diamond_sword": {},
                                        "make_diamond_armour": {},
                                    },
                                    "place_furnace": {
                                        "make_iron_sword": {},
                                        "make_iron_armour": {},
                                    },
                                }
                            },
                            "Collect Coal": {},
                        },
                    }
                },
                "make_wood_sword": {},
            }
        },
        "Eat Cow": {},
        "Collect Sapling": {"place_plant": {"Eat Plant": {}}},
        "Collect Drink": {},
        "Defeat Zombie": {},
        "Defeat Skeleton": {},
        "Wake Up": {},
        "Make Arrow": {},
        "Make Torch": {"Place Torch": {}},
        "Enter Gnomish Mines": {"Defeat Gnome Warrior": {}, "Defeat Gnome Archer": {}},
        "Enter Dungeon": {},
        "Enter Sewers": {},
        "Enter Vault": {},
        "Enter Troll Mines": {"Defeat Troll": {}},
        "Enter Fire Realm": {"Defeat Fire Elemental": {}, "Defeat Pigman": {}},
        "Enter Ice Realm": {"Defeat Frost Troll": {}, "Defeat Ice Elemental": {}},
        "Enter Graveyard": {"Damage Necromancer": {"Defeat Necromancer": {}}},
        "Defeat Orc Soldier": {},
        "Defeat Orc Mage": {},
        "Defeat Lizard": {},
        "Defeat Kobold": {},
        "Defeat Deep Thing": {},
        "Eat Bat": {},
        "Eat Snail": {},
        "Find Bow": {"Fire Bow": {}},
        "Collect Sapphire": {"Learn Iceball": {"Cast Iceball": {}}},
        "Collect Ruby": {"Learn Fireball": {"Cast Fireball": {}}},
        "Open Chest": {},
        "Drink Potion": {},
        "Enchant Sword": {},
        "Enchant Armour": {},
        "Defeat Knight": {},
        "Defeat Archer": {},
    },
}
craftax_full_action_dict = {
    "noop": 0,
    "left": 1,
    "right": 2,
    "up": 3,
    "down": 4,
    "do": 5,
    "sleep": 6,
    "place_stone": 7,
    "place_table": 8,
    "place_furnace": 9,
    "place_plant": 10,
    "make_wood_pickaxe": 11,
    "make_stone_pickaxe": 12,
    "make_iron_pickaxe": 13,
    "make_wood_sword": 14,
    "make_stone_sword": 15,
    "make_iron_sword": 16,
    "rest": 17,
    "descend": 18,
    "ascend": 19,
    "make_diamond_pickaxe": 20,
    "make_diamond_sword": 21,
    "make_iron_armour": 22,
    "make_diamond_armour": 23,
    "shoot_arrow": 24,
    "make_arrow": 25,
    "cast_fireball": 26,
    "cast_iceball": 27,
    "place_torch": 28,
    "drink_potion_red": 29,
    "drink_potion_green": 30,
    "drink_potion_blue": 31,
    "drink_potion_pink": 32,
    "drink_potion_cyan": 33,
    "drink_potion_yellow": 34,
    "read_book": 35,
    "enchant_sword": 36,
    "enchant_armour": 37,
    "make_torch": 38,
    "level_up_dexterity": 39,
    "level_up_strength": 40,
    "level_up_intelligence": 41,
    "enchant_bow": 42,
}
craftax_game_tips = [
    "Explore dungeons, mine resources, craft items, and fight enemies to progress through the game.",
    "Manage your health, hunger, thirst, energy, and mana to survive. Replenish them by eating, drinking, sleeping, and resting.",
    "Find and use ladders to descend to the next level after defeating 8 creatures on each floor.",
    "Start in the Overworld by gathering wood, crafting tools, and preparing for dungeon exploration.",
    "Craft tables and furnaces to create more advanced tools and weapons.",
    "Use the bow and arrows for ranged attacks, and craft arrows using wood and stone.",
    "Assign experience points to upgrade Dexterity, Strength, or Intelligence attributes.",
    "Experiment with different colored potions to discover their effects, which change each game.",
    "Use torches in dark levels to navigate and explore safely.",
    "Craft and wear armor to reduce damage from enemies.",
    "Learn and use spells for ranged magical attacks, especially against enemies resistant to physical damage.",
    "Enchant weapons and armor using gemstones and mana for elemental bonuses.",
    "Mine valuable ores like diamonds, sapphires, and rubies for crafting and enchanting.",
    "Adapt your strategy for each unique floor, considering the environment and enemy types.",
    "Build bridges across lava using stone in the Fire Realm.",
    "Prepare for the absence of food or water on certain levels by stocking up beforehand.",
    "Defeat the Necromancer in the final Graveyard level by surviving waves of enemies and attacking during vulnerable states.",
]



from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel

from apropos.src.core.lms.helpers import LLM


class Topic(BaseModel):
    topic_name: str
    topic_template: str
    instructions_fields: Dict[str, str]
    input_fields: List[str]


class SystemMessage(BaseModel):
    premise: List[
        Topic
    ] = []  # Assuming 'premise' and 'objective' are the keys you need
    objective: List[Topic] = []
    constraints: List[Topic] = []  # Add this if 'constraints' is also a key you need


class UserMessage(BaseModel):
    user: List[Topic] = []  # Assuming 'user' is the key for user topics


class Demonstration(BaseModel):
    inputs: Union[Dict, str]
    outputs: Union[Dict, str]
    prompt: str
    gold_outputs: Optional[str]
    annotation: Optional[str]

    def to_string(self):
        return f"Input: {self.inputs}\nOutput: {self.outputs}\nPrompt: {self.prompt}\nGold Output: {self.gold_outputs}\nAnnotation: {self.annotation}"


class PromptTemplate(BaseModel):
    name: str
    system: SystemMessage
    user: UserMessage
    response_type: Literal["pydantic", "str"]
    response_model_scheme: Optional[Dict]
    demonstrations: List[Demonstration]

    def to_dict(self):
        return {
            "name": self.name,
            "system": self.system.dict(),
            "user": self.user.dict(),
            "response_type": self.response_type,
            "response_model_scheme": self.response_model_scheme,
            "demonstrations": [d.dict() for d in self.demonstrations],
        }

    def get_input_fields(self):
        input_fields = []
        for topic in self.system.premise:
            input_fields += topic.input_fields
        for topic in self.system.objective:
            input_fields += topic.input_fields
        for topic in self.system.constraints:
            input_fields += topic.input_fields
        for topic in self.user.user:
            input_fields += topic.input_fields
        return input_fields

    def get_line_by_key(self, key: str):
        def search_segment_for_key(segment: List[Topic]):
            for topic in segment:
                if key in topic.instructions_fields.keys():
                    return topic.instructions_fields[key]
                if key in topic.input_fields:
                    return topic.input_fields[key]
            return None

        for source in [
            self.system.premise,
            self.system.objective,
            self.system.constraints,
            self.user.user,
        ]:
            result = search_segment_for_key(source)
            if result:
                return result
        return None

    def set_line_by_key(self, key: str, value: str):
        def search_segment_for_key(segment: List[Topic], key: str, value: str):
            for topic in segment:
                if key in topic.instructions_fields.keys():
                    topic.instructions_fields[key] = value
                if key in topic.input_fields:
                    topic.input_fields[key] = value
            return segment

        self.system.premise = search_segment_for_key(self.system.premise, key, value)
        self.system.objective = search_segment_for_key(
            self.system.objective, key, value
        )
        self.system.constraints = search_segment_for_key(
            self.system.constraints, key, value
        )
        self.user.user = search_segment_for_key(self.user.user, key, value)

    def to_markdown(self):
        markdown = f"# Template: {self.name}\n\n"
        markdown += "## System\n\n"
        for topic in self.system.premise:
            markdown += f"### {topic.topic_name}\n\n"
            markdown += f"**Template**: {topic.topic_template}\n\n"
            markdown += f"**Instructions Fields**: {topic.instructions_fields}\n\n"
            markdown += f"**Input Fields**: {topic.input_fields}\n\n"
        markdown += "## User\n\n"
        for topic in self.user.user:
            markdown += f"### {topic.topic_name}\n\n"
            markdown += f"**Template**: {topic.topic_template}\n\n"
            markdown += f"**Instructions Fields**: {topic.instructions_fields}\n\n"
            markdown += f"**Input Fields**: {topic.input_fields}\n\n"
        return markdown

    def compile(
        self, inputs: Dict[str, str], custom_instructions_fields: Dict[str, str] = {}
    ):
        system = ""
        for topic in self.system.premise:
            template = topic.topic_template
            for field in topic.input_fields:
                if field not in template:
                    continue

                template = template.replace(
                    f"{field}", inputs[field] if inputs[field] else ""
                )
            for field in topic.instructions_fields:
                if field not in template:
                    continue
                if field in custom_instructions_fields:
                    template = template.replace(
                        f"{field}", custom_instructions_fields[field]
                    )
                else:
                    template = template.replace(
                        f"{field}", topic.instructions_fields[field]
                    )
            system += template + "\n\n"
        # system+="\n\nObjective:\n\n"
        for topic in self.system.objective:
            template = topic.topic_template
            for field in topic.input_fields:
                if field not in template:
                    continue
                template = template.replace(
                    f"{field}", inputs[field] if inputs[field] else ""
                )
            for field in topic.instructions_fields:
                if field not in template:
                    continue
                if field in custom_instructions_fields:
                    template = template.replace(
                        f"{field}", custom_instructions_fields[field]
                    )
                else:
                    template = template.replace(
                        f"{field}", topic.instructions_fields[field]
                    )
            system += template + "\n\n"

        for topic in self.system.constraints:
            template = topic.topic_template
            for field in topic.input_fields:
                if field not in template:
                    continue
                template = template.replace(
                    f"{field}", inputs[field] if inputs[field] else ""
                )
            for field in topic.instructions_fields:
                if field not in template:
                    continue
                if field in custom_instructions_fields:
                    template = template.replace(
                        f"{field}", custom_instructions_fields[field]
                    )
                else:
                    template = template.replace(
                        f"{field}", topic.instructions_fields[field]
                    )
            system += template + "\n\n"
        if len(self.demonstrations) > 0:
            system += "\n\n Demonstrations of Success:\n\n"
            for demo in self.demonstrations:
                annotation_snippet = (
                    f"Annotation: {demo.annotation}\n" if demo.annotation else ""
                )
                system += f"Input: {demo.inputs}\nOutput: {demo.outputs}\n{annotation_snippet}\n"

        user = ""
        for topic in self.user.user:
            template = topic.topic_template
            for field in topic.input_fields:
                if field not in template:
                    continue
                template = template.replace(
                    f"{field}", inputs[field] if inputs[field] else ""
                )
            for field in topic.instructions_fields:
                if field not in template:
                    continue
                if field in custom_instructions_fields:
                    template = template.replace(
                        f"{field}", custom_instructions_fields[field]
                    )
                else:
                    template = template.replace(
                        f"{field}", topic.instructions_fields[field]
                    )
            user += template + "\n\n"
        return system, user

    def run(
        self,
        inputs: Dict[str, str],
        lm: LLM,
        custom_instructions_fields: Dict[str, str] = {},
        response_model: Optional[Any] = None,
    ):
        system, user = self.compile(inputs, custom_instructions_fields)
        if response_model:
            response = lm.respond(system, user, response_model)
        else:
            response = lm.respond(system, user)
        return response

    async def arun(
        self,
        inputs: Dict[str, str],
        lm: LLM,
        custom_instructions_fields: Dict[str, str] = {},
        response_model: Optional[Any] = None,
    ):
        system, user = self.compile(inputs, custom_instructions_fields)
        if response_model:
            response = await lm.async_respond(
                system_prompt=system, user_prompt=user, response_model=response_model
            )
        else:
            response = await lm.async_respond(system_prompt=system, user_prompt=user)
        assert isinstance(response, str) or isinstance(
            response, BaseModel
        ), f"Response: {response}"
        return response



import asyncio

from apropos.src.bench.base import Benchmark
from apropos.src.bench.hendryks_math.dags.plan_execute import (
    hendryks_math_plan_execute_example,
)
from apropos.src.bench.hendryks_math.dags.single_step import (
    hendryks_math_single_step_example,
)
from apropos.src.bench.hendryks_math.main import HendryksMath_Benchmark
from apropos.src.core.optimizers.baselines.bffsrs import BreadthFirstRandomSearch_DAG
from apropos.src.core.programs.dag import LM_DAG


async def main(
    benchmark: Benchmark,
    plan_execute_dag: LM_DAG,
    single_step_dag: LM_DAG,
    bffsrs_config: dict,
):
    bffsrs = BreadthFirstRandomSearch_DAG(
        student_program=plan_execute_dag,
        teacher_program=plan_execute_dag,
        dataset_handler=benchmark,
        cfg=bffsrs_config,
    )
    print("Optimizing demonstrations...")
    optimized_program_with_demos = await bffsrs.optimize_demonstrations()

    print("Evaluating programs...")
    baseline_program_scores, _ = await benchmark.score_dag(
        plan_execute_dag, n=100, verbose=True, split="test", patches=["A", "B"]
    )
    optimized_program_scores, _ = await benchmark.score_dag(
        optimized_program_with_demos,
        n=100,
        verbose=True,
        split="test",
        patches=["A", "B"],
    )

    single_step_program_scores, _ = await benchmark.score_dag(
        single_step_dag, n=100, verbose=True, split="test", patches=["A", "B"]
    )

    print(
        f"Unoptimized Plan-Execute program score: {sum(baseline_program_scores)/len(baseline_program_scores)}"
    )
    print(
        f"Optimized Plan-Execute program score: {sum(optimized_program_scores)/len(optimized_program_scores)}"
    )
    print(
        f"Baseline Single-Step program score: {sum(single_step_program_scores)/len(single_step_program_scores)}"
    )

    pass


if __name__ == "__main__":
    benchmark = HendryksMath_Benchmark()
    plan_execute_dag = hendryks_math_plan_execute_example(
        model_names=["claude-3-haiku-20240307"] * 2
    )
    baseline_single_step_program = hendryks_math_single_step_example(
        model_name="claude-3-haiku-20240307"
    )
    bffsrs_config = {
        "optimization": {
            "combination_size": 3,
            "n_iterations": 10,
            "validation_size": 30,
            "test_size": 100,
            "program_search_parallelization_factor": 5,
        },
        "bootstrapping": {
            "patches": ["A", "B"],
            "n_questions": 100,
        },
        "verbose": True,
    }
    asyncio.run(
        main(
            benchmark=benchmark,
            plan_execute_dag=plan_execute_dag,
            single_step_dag=baseline_single_step_program,
            bffsrs_config=bffsrs_config,
        )
    )
    # After 10 iterations
    # Unoptimized Plan-Execute program score: 0.13
    # Optimized Plan-Execute program score: 0.28
    # Baseline Single-Step program score: 0.24



import asyncio
import os
from typing import Dict, List, Literal

import numpy as np
from pydantic import BaseModel

from apropos.src.bench.crafter.game_dynamics import harness
from apropos.src.bench.crafter.game_information import (
    action_dict,
    crafter_game_rules,
    crafter_game_tips,
)
import base64
from io import BytesIO
from PIL import Image
from apropos.src.core.lms.cost import CostMonitor
from apropos.src.core.lms.helpers import LLM


class ReAct(BaseModel):
    reasoning: str
    actions: List[str]


class SimpleReActLanguageAgent:
    obs_history: List[Dict]
    react_history: List[Dict]
    lm: LLM
    cost_monitor: CostMonitor

    def __init__(self, lm: LLM, cost_monitor: CostMonitor):
        self.obs_history = []
        self.react_history = []
        self.lm = lm
        self.cost_monitor = cost_monitor

    def render_history(self):
        react_history = [
            f"<{i} reasoning step(s) in the past>{item}</{i} reasoning step(s) in the past>"
            for i, item in enumerate(reversed(self.react_history[-5:]), 1)
        ]
        obs_history = [
            f"<{i} environment step(s) in the past>{item}</{i} environment step(s) in the past>"
            for i, item in enumerate(reversed(self.obs_history[-5:]), 1)
        ]
        return "\n".join(react_history), "\n".join(obs_history)

    async def get_actions(self):
        system_message = f"""
# Premise
You're playing the game of Crafter.
Here is some information about this setting
<Crafter Information>
<Rules>
{crafter_game_rules}
</Rules>
<Tips>
{crafter_game_tips}
</Tips>
<Actions Available>
{list(action_dict.values())}
</Actions Available>
You'll be given your past actions/thoughts, along with recent raw observations from the environment

# Objective
Please choose 4-7 actions to take next in the game, and provide a thought justifying the long-term end you'd like to achieve with these actions.

# Constraints
- If your previous actions worked toward a valuable goal, continue in that direction unless you have a good reason to change course.
- If there's no important thought to share, don't share anything. "No update" is a valid and the most common response.
- Render your actions by name. All rendered names must match the names in the "Actions Available" section.
- If you attempt an action that requires certain conditions to be met without meeting those conditions, the action will default to a no-op.
"""

        react_history, obs_history = self.render_history()
        user_message = f"""
# Recent Actions / Thoughts
{react_history}
# Recent Observations
{obs_history}

Your next actions / thought: """
        react_step = await self.lm.async_respond(
            system_prompt=system_message,
            user_prompt=user_message,
            response_model=ReAct,
        )
        # if len(self.react_history) > 4:
        #     print("System Message", system_message)
        #     print("User Message", user_message)
        #     print("Agent Response", react_step)
        #     raise ValueError("All actions are noops")

        illegal_actions = [
            action
            for action in react_step.actions
            if action not in action_dict.values()
        ]
        legal_actions = [
            action for action in react_step.actions if action in action_dict.values()
        ]
        react_info = react_step.dict()
        react_info["errors"] = {
            "illegal_actions": illegal_actions,
        }
        self.react_history.append(react_info)
        self.cost_monitor.update_token_counts(
            system_message, user_message, str(react_step.dict())
        )
        return [
            {v: k for k, v in action_dict.items()}[action_name]
            for action_name in legal_actions
        ]

    async def add_observations(self, observations: List[Dict]):
        for observation in observations:
            self.obs_history.append(observation["text_obs"][0].default_readout())


class SimpleVisionLanguageAgent:
    obs_history: List[Dict]
    react_history: List[Dict]
    lm: LLM
    cost_monitor: CostMonitor

    def __init__(self, lm: LLM, cost_monitor: CostMonitor):
        self.obs_history = []
        self.react_history = []
        self.lm = lm
        self.cost_monitor = cost_monitor

    def render_history(self):
        react_history = [
            f"<{i} reasoning step(s) in the past>{item}</{i} reasoning step(s) in the past>"
            for i, item in enumerate(reversed(self.react_history[-5:]), 1)
        ]
        obs_history = [
            f"<{i} environment step(s) in the past>{item['text']}</{i} environment step(s) in the past>"
            for i, item in enumerate(reversed(self.obs_history[-5:]), 1)
        ]
        return "\n".join(react_history), "\n".join(obs_history)

    async def get_actions(self):
        subset_actions = list(action_dict.values())
        # subset_actions = ['Move West', 'Move East', 'Move North', 'Move South', 'Do']#, 'Place Table'
        system_message = f"""
# Premise
You're playing the game of Crafter.
Here is some information about this setting
<Crafter Information>
<Rules>
{crafter_game_rules}
</Rules>
<Tips>
{crafter_game_tips}
</Tips>
<Actions Available>
{subset_actions}
</Actions Available>
You'll be given your past actions/thoughts, along with recent raw observations from the environment.
The images below are the most recent observations from the environment. The first image is the most recent observation, and the last image is the oldest observation.

# Objective
Please choose 4-7 actions to take next in the game, and provide a thought justifying the long-term end you'd like to achieve with these actions.

# Constraints
- If your previous actions worked toward a valuable goal, continue in that direction unless you have a good reason to change course.
- If there's no important thought to share, don't share anything. "No update" is a valid and the most common response.
- Render your actions by name. All rendered names must match the names in the "Actions Available" section.
- If you attempt an action that requires certain conditions to be met without meeting those conditions, the action will default to a no-op.
- Don't choose no-op as an action.
- Don't choose actions that require conditions that have not been met.
"""

        react_history, obs_history = self.render_history()
        user_message = f"""
# Recent Actions / Thoughts
{react_history}
# Recent Observations
{[obs_history_datum for obs_history_datum in obs_history]}
Your next actions / thought: """
        visual_recent_obs = [obs_history["image"] for obs_history in self.obs_history]
        visual_recent_obs_bytes = []
        for obs in visual_recent_obs:
            if isinstance(obs, bytes):
                img_array = np.frombuffer(obs, dtype=np.uint8)
                img_array = img_array.reshape(
                    (64, 64, 3)
                )  # Assuming 64x64x3 image size
                img = Image.fromarray(img_array)
            else:
                img = Image.fromarray(obs)

            buffered = BytesIO()
            img.convert("RGB").save(buffered, format="PNG")
            img_str = base64.b64encode(buffered.getvalue()).decode()
            visual_recent_obs_bytes.append(img_str)
        react_step = await self.lm.async_respond(
            system_prompt=system_message,
            user_prompt=user_message,
            images_bytes=list(reversed(visual_recent_obs_bytes[-10:])),
            response_model=ReAct,
        )
        illegal_actions = [
            action
            for action in react_step.actions
            if action not in action_dict.values()
        ]
        legal_actions = [
            action for action in react_step.actions if action in action_dict.values()
        ]

        react_info = react_step.dict()
        react_info["errors"] = {
            "illegal_actions": illegal_actions,
        }
        self.react_history.append(react_info)
        self.cost_monitor.update_token_counts(
            system_message, user_message, str(react_step.dict())
        )
        return [
            {v: k for k, v in action_dict.items()}[action_name]
            for action_name in legal_actions
        ]

    async def add_observations(self, observations: List[Dict]):
        for observation in observations:
            self.obs_history.append(
                {
                    "text": observation["text_obs"][0].visual_readout(),
                    "image": observation["vis_obs"],
                }
            )


def hafner_scoring_function_modified(achievements_probs: Dict):
    probs = list(achievements_probs.values())
    return sum([np.log(1 + prob) for prob in probs])


# Log token count to get approx cost of rollout.
async def score_rollouts(
    k_steps=10,
    n_rollouts=10,
    verbose=False,
    model_name="gpt-4o-mini",
    modality: Literal["text", "vision"] = "text",
):
    core_lm = LLM(model_name)
    cost_monitor = CostMonitor(model_name=core_lm.model_name)
    if modality == "text":
        agent = SimpleReActLanguageAgent(lm=core_lm, cost_monitor=cost_monitor)
    elif modality == "vision":
        agent = SimpleVisionLanguageAgent(lm=core_lm, cost_monitor=cost_monitor)

    async def single_rollout():
        # try:
        achievements_log = await harness(agent, k_steps=k_steps, verbose=verbose)
        # except Exception as e:
        #     print(f"Error in Rollout: {e}")
        #     return
        achievements = list({k: v for k, v in achievements_log.items() if v > 0}.keys())
        action_density = len(agent.obs_history) / len(agent.react_history)
        price, n_tokens = cost_monitor.final_cost()
        if verbose and action_density < 2:
            print(f"Warning - Low Action Density of {action_density}")
        if verbose:
            print(f"Rollout price ${price:.2f} for {n_tokens} tokens")
        return achievements, price

    achivements_and_price_by_rollout = await asyncio.gather(
        *[single_rollout() for _ in range(n_rollouts)]
    )
    achivements_and_price_by_rollout = [
        a_and_p for a_and_p in achivements_and_price_by_rollout if a_and_p is not None
    ]
    total_price = sum([price for _, price in achivements_and_price_by_rollout])
    achievements_probs = {
        k: len(
            [
                achievements
                for achievements, _ in achivements_and_price_by_rollout
                if k in achievements
            ]
        )
        / len(achivements_and_price_by_rollout)
        for k in set(
            [
                achievement
                for achievements, _ in achivements_and_price_by_rollout
                for achievement in achievements
            ]
        )
    }
    return (
        hafner_scoring_function_modified(achievements_probs),
        achievements_probs,
        total_price,
    )


if __name__ == "__main__":
    # gpt-4o-2024-08-06
    # gpt-4o-mini-2024-07-18
    import jax

    # jax.config.update('jax_disable_jit', True)
    hafner_score, achievement_probs, total_price = asyncio.run(
        score_rollouts(
            k_steps=300,
            n_rollouts=1,
            verbose=True,
            model_name="gpt-4o-mini-2024-07-18",
            modality="text",
        )  # hermes-3-llama-3.1-405b-fp8-128k
    )
    print("Agent got a normalized Crafter score of", hafner_score)
    print("Achievement Probabilities:")
    for k, v in achievement_probs.items():
        print(f"{k}: {v:.3f}")
    print(f"Total price of experiment: ${total_price:.2f}")
    # TODO: add a way to price images
    # Up is left
    # Left is down
    # Down is Right
    # Right is up



import asyncio
from typing import Dict, Literal

import numpy as np

from apropos.src.bench.craftax.game_dynamics import harness
from apropos.src.bench.craftax.react import SimpleReActLanguageAgent
from apropos.src.core.lms.cost import CostMonitor
from apropos.src.core.lms.helpers import LLM


def hafner_scoring_function_modified(achievements_probs: Dict):
    probs = list(achievements_probs.values())
    return sum([np.log(1 + prob) for prob in probs])


# Log token count to get approx cost of rollout.
async def score_rollouts(
    k_steps=10,
    base_seed=0,
    n_rollouts=10,
    verbose=False,
    model_name="gpt-4o-mini",
    modality: Literal["text", "vision"] = "text",
    mode: Literal["craftax_classic", "craftax_full"] = "craftax_classic",
):
    core_lm = LLM(model_name)
    cost_monitor = CostMonitor(model_name=core_lm.model_name)
    if modality == "text":
        agent = SimpleReActLanguageAgent(
            lm=core_lm, cost_monitor=cost_monitor, mode=mode
        )
    elif modality == "vision":
        raise NotImplementedError("Vision not implemented yet")

    async def single_rollout(seed):
        achievements_log = await harness(
            agent, seed=base_seed + seed, k_steps=k_steps, verbose=verbose, mode=mode
        )
        achievements = list({k: v for k, v in achievements_log.items() if v > 0}.keys())
        print("Achievements", achievements)
        action_density = len(agent.obs_history) / len(agent.react_history)
        price, n_tokens = cost_monitor.final_cost()
        if verbose and action_density < 2:
            print(f"Warning - Low Action Density of {action_density}")
        if verbose:
            print(f"Rollout price ${price:.2f} for {n_tokens} tokens")
        return achievements, price

    achivements_and_price_by_rollout = await asyncio.gather(
        *[single_rollout(seed=_) for _ in range(n_rollouts)]
    )
    achivements_and_price_by_rollout = [
        a_and_p for a_and_p in achivements_and_price_by_rollout if a_and_p is not None
    ]
    total_price = sum([price for _, price in achivements_and_price_by_rollout])
    achievements_probs = {
        k: len(
            [
                achievements
                for achievements, _ in achivements_and_price_by_rollout
                if k in achievements
            ]
        )
        / len(achivements_and_price_by_rollout)
        for k in set(
            [
                achievement
                for achievements, _ in achivements_and_price_by_rollout
                for achievement in achievements
            ]
        )
    }
    return (
        hafner_scoring_function_modified(achievements_probs),
        achievements_probs,
        total_price,
    )


if __name__ == "__main__":
    # gpt-4o-2024-08-06
    # gpt-4o-mini-2024-07-18
    # meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo
    model_name = (
        "gpt-4o-mini-2024-07-18"  # "ft:gpt-4o-mini-2024-07-18:basis:fbc-0:9yMCGTnx"
    )
    mode = "craftax_classic"
    hafner_score, achievement_probs, total_price = asyncio.run(
        score_rollouts(
            k_steps=300,
            n_rollouts=10,
            base_seed=1000,
            verbose=False,
            model_name=model_name,
            modality="text",
            mode=mode,
        )  # hermes-3-llama-3.1-405b-fp8-128k
    )
    print("Agent got a normalized Crafter score of", hafner_score, "on mode: ", mode)
    print("Achievement Probabilities:")
    for k, v in achievement_probs.items():
        print(f"{k}: {v:.3f}")
    print(
        f"Total price of experiment: ${total_price:.2f}"
    )  # TODO: add a way to price images

    # Info for FT experiments
    # Agent got a normalized Crafter score of 2.824573777305732 on mode:  craftax_classic
    # Achievement Probabilities:
    # Collect Wood: 1.000
    # Make Wood Pickaxe: 0.400
    # Collect Drink: 0.900
    # Place Plant: 0.100
    # Place Table: 1.000
    # Collect Stone: 0.200
    # Collect Sapling: 0.200

    # claude-3-5-sonnet-20240620 @ classic - Aug 19, 2024 @ base seed 0
    # Josh note - a bit shocked, it typically does much better than gpt-4o. Going to use discretion and not add to the leaderboard.
    # Agent got a normalized Crafter score of 3.7285323200296063 on mode:  craftax_classic
    # Achievement Probabilities:
    # Make Wood Sword: 0.200
    # Make Wood Pickaxe: 0.600
    # Wake Up: 0.600
    # Collect Coal: 0.400
    # Collect Stone: 0.400
    # Place Table: 0.800
    # Collect Wood: 1.000
    # Collect Drink: 0.600
    # Make Stone Pickaxe: 0.200

    # gpt-4o-2024-08-06 @ classic - Aug 19, 2024 @ base seed 0
    # Achievements ['Collect Wood', 'Place Table', 'Collect Drink', 'Make Wood Pickaxe', 'Make Wood Sword', 'Collect Stone', 'Wake Up', 'Place Furnace', 'Collect Coal']
    # Achievements ['Collect Wood', 'Place Table', 'Collect Drink', 'Make Wood Sword', 'Wake Up']
    # Achievements ['Collect Wood', 'Place Table', 'Collect Drink', 'Make Wood Pickaxe', 'Collect Stone', 'Wake Up', 'Place Furnace', 'Collect Coal']
    # Achievements ['Collect Wood', 'Wake Up']
    # Achievements ['Collect Wood', 'Place Table', 'Make Wood Pickaxe', 'Make Wood Sword', 'Collect Stone', 'Wake Up']
    # Agent got a normalized Crafter score of 4.527040016247378/15.2492 on mode:  craftax_classic
    # Achievement Probabilities:
    # Collect Coal: 0.400
    # Collect Wood: 1.000
    # Make Wood Sword: 0.600
    # Make Wood Pickaxe: 0.600
    # Collect Drink: 0.600
    # Place Furnace: 0.400
    # Collect Stone: 0.600
    # Place Table: 0.800
    # Wake Up: 1.000

    # gpt-4o-mini-2024-07-18 @ classic - Aug 19, 2024 @ base seed 0
    # Achievements ['Collect Wood', 'Place Table']
    # Achievements ['Collect Wood', 'Collect Drink']
    # Achievements ['Collect Wood', 'Place Table', 'Collect Sapling', 'Make Wood Pickaxe', 'Collect Stone', 'Collect Coal']
    # Achievements ['Collect Wood', 'Place Table', 'Collect Drink']
    # Achievements ['Collect Wood', 'Place Table', 'Make Wood Pickaxe', 'Collect Stone', 'Collect Coal']
    # Agent got a normalized Crafter score of 2.8091443487408707/15.2492
    # Achievement Probabilities:
    # Collect Coal: 0.400
    # Make Wood Pickaxe: 0.400
    # Collect Stone: 0.400
    # Collect Sapling: 0.200
    # Place Table: 0.800
    # Collect Drink: 0.400
    # Collect Wood: 1.000

    # gpt-4o-mini-2024-07-18 @ classic - Aug 19, 2024 @ base seed 20 5 x 300
    # ['Collect Wood', 'Place Table', 'Make Wood Pickaxe', 'Collect Stone']
    # ['Collect Wood', 'Place Table', 'Collect Drink']
    # ['Collect Wood', 'Place Table', 'Collect Drink', 'Make Wood Pickaxe', 'Collect Stone']
    # ['Collect Wood', 'Place Table', 'Collect Drink', 'Make Wood Pickaxe']
    # ['Collect Wood', 'Place Table', 'Eat Cow', 'Make Wood Pickaxe']
    # Agent got a normalized Crafter score of 2.962878448682913 on mode:  craftax_classic
    # Achievement Probabilities:
    # Place Table: 1.000
    # Collect Stone: 0.400
    # Eat Cow: 0.200
    # Collect Wood: 1.000
    # Collect Drink: 0.600
    # Make Wood Pickaxe: 0.800

    # ft:gpt-4o-mini-2024-07-18:basis:fbc-0:9yMCGTnx @ classic - Aug 19, 2024 @ base seed 1000 5 x 300
    # FT on 100 filtered trajectories, 5 steps before reward (600 messages total)
    # Agent got a normalized Crafter score of 2.9628784486829125 on mode:  craftax_classic
    # Place Table: 1.000
    # Collect Wood: 1.000
    # Make Wood Pickaxe: 0.800
    # Collect Drink: 0.400
    # Collect Coal: 0.200
    # Collect Stone: 0.600

    # ft:gpt-4o-mini-2024-07-18:basis:fbc-full:9yQ7idkM @ classic - Aug 19, 2024 @ base seed 20 5 x 300

    # deepseek-chat @ classic - Aug 19, 2024
    # Agent got a normalized Crafter score of 2.662545258677573/15.2492
    # Achievement Probabilities:
    # Collect Stone: 0.200
    # Collect Drink: 0.800
    # Make Stone Pickaxe: 0.200
    # Wake Up: 0.600
    # Place Table: 0.200
    # Collect Wood: 1.000
    # Make Wood Pickaxe: 0.200
    # Collect Sapling: 0.200

    # gpt-4o-mini-2024-07-18 @ full - Aug 19, 2024
    # Agent got a normalized Crafter score of 1.3046503720793805 / 45.747
    # Achievement Probabilities:
    # collect_wood: 0.600
    # place_table: 0.200
    # collect_drink: 0.600
    # collect_sapling: 0.200

    # Agent got a normalized Crafter score of 1.6943671232194055
    # Achievement Probabilities:
    # collect_sapling: 0.200
    # collect_drink: 0.400
    # place_table: 0.800
    # collect_wood: 0.800

    # flash @ classic - Aug 19, 2024

    # flash 8b 0.988/ 45.747
    # 0.6523
    # Collect Wood: 0.600
    # Collect Drink: 0.200

    # llama 3 8b 0.77010 / 45.747



from typing import Literal
from craftaxlm import CraftaxACI, CraftaxClassicACI


async def harness(
    agent,
    k_steps,
    seed=0,
    verbose=False,
    mode: Literal["craftax_classic", "craftax_full"] = "craftax_classic",
):
    if mode == "craftax_classic":
        aci = CraftaxClassicACI(seed=seed, verbose=verbose)
    elif mode == "craftax_full":
        aci = CraftaxACI(seed=seed, verbose=verbose)
    else:
        raise ValueError("Invalid Env mode")
    done = False
    step_infos = [aci.starting_obs]
    while not done and k_steps > 0:
        await agent.add_observations(step_infos)
        actions = await agent.get_actions()
        k_steps -= len(actions)
        step_infos, rewards, done = aci.multistep(actions)
        if done:
            break
    achievements = aci.terminate()
    return achievements


if __name__ == "__main__":
    pass



from typing import List, Tuple
import re
import sys

# cv2, pyplot
module_to_package_mapping = {
    "cv2": "opencv-python",
    "sklearn": "scikit-learn",
    "dateutil": "python-dateutil",
    "PIL": "Pillow",
}
tabu = ["urllib", "urllib.request", "urllib.parse","mpl_toolkits"]


def get_imports(code: str) -> Tuple[List[str], str]:
    from_imports = re.findall(r"from ([\w.]+) import [\w, ]+", code)
    head_imports = re.findall(r"^import (\w+)(?!\s+as)$", code, re.MULTILINE)
    module_head_imports = re.findall(r"^import ([\w.]+)(?!\s+as)$", code, re.MULTILINE)
    head_imports_with_alias = re.findall(r"import ([\w.]+) as (\w+)", code)
    imports_snippet = "\n".join(
        ["import " + imp for imp in head_imports]
        + [f"import {imp} as {alias}" for imp, alias in head_imports_with_alias]
        + [
            line.strip()
            for line in code.split("\n")
            if line.strip().startswith("from ")
        ]
    )
    standard_libs = set(sys.stdlib_module_names)
    all_unique_imports = set()
    for imp in (
        head_imports + module_head_imports + [imp for imp, _ in head_imports_with_alias]
    ):
        parts = imp.split(".")
        for i in range(len(parts)):
            package = ".".join(parts[: i + 1])
            if package not in standard_libs:
                all_unique_imports.add(package)
                break
    for package in from_imports:
        parts = package.split(".")
        for i in range(len(parts)):
            subpackage = ".".join(parts[: i + 1])
            if subpackage not in standard_libs:
                all_unique_imports.add(subpackage)
                break
    return [
        module_to_package_mapping.get(v, v) for v in all_unique_imports if v not in tabu
    ], imports_snippet


def test_get_imports():
    test_cases = [
        (
            "import cv2\nimport matplotlib.pyplot as plt",
            (
                ["opencv-python", "matplotlib"],
                "import cv2\nimport matplotlib.pyplot as plt",
            ),
        ),
        (
            "from numpy import array\nfrom scipy import stats",
            (["numpy", "scipy"], "from numpy import array\nfrom scipy import stats"),
        ),
        (
            "import os\nfrom datetime import datetime",
            ([], "import os\nfrom datetime import datetime"),
        ),
        (
            "import pandas as pd\nfrom sklearn.model_selection import train_test_split",
            (
                ["pandas", "scikit-learn"],
                "import pandas as pd\nfrom sklearn.model_selection import train_test_split",
            ),
        ),
        (
            "import tensorflow as tf\nfrom keras.layers import Dense",
            (
                ["tensorflow", "keras"],
                "import tensorflow as tf\nfrom keras.layers import Dense",
            ),
        ),
        (
            "import dateutil.parser\nfrom datetime import datetime",
            (
                ["python-dateutil"],
                "import dateutil.parser\nfrom datetime import datetime",
            ),
        ),
    ]

    failures = 0
    successes = 0

    for code, expected_output in test_cases:
        packages, imports_snippet = get_imports(code)
        if sorted(expected_output[0]) != sorted(
            packages
        ):  # or imports_snippet != expected_output[1]
            failures += 1
            print(f"Test case failed:")
            print(f"Input: {code}")
            print(f"Expected packages: {expected_output[0]}, got: {packages}")
            print(f"Expected imports snippet: {expected_output[1]}")
            print(f"Got imports snippet: {imports_snippet}")
            print()
        else:
            successes += 1

    print(f"Test results: {successes} passed, {failures} failed")


if __name__ == "__main__":
    test_get_imports()



import os
from datetime import datetime
from typing import Dict, List, Optional, Type

import loguru
from pydantic import BaseModel

from apropos.src.core.lms.vendors.anthropic_api import AnthropicAPIProvider
from apropos.src.core.lms.vendors.deepmind_api import DeepmindAPIProvider
from apropos.src.core.lms.vendors.deepseek_api import DeepSeekAPIProvider
from apropos.src.core.lms.vendors.groq_api import GroqAPIProvider
from apropos.src.core.lms.vendors.lambda_api import LambdaAPIProvider
from apropos.src.core.lms.vendors.ollama_api import OllamaAPIProvider
from apropos.src.core.lms.vendors.openai_api import OpenAIAPIProvider
from apropos.src.core.lms.vendors.together_api import TogetherAPIProvider

logger = loguru.logger


def log_backoff(details):
    pass


def build_messages(
    sys_msg: str,
    user_msg: str,
    images_bytes: List = [],
    model_name: Optional[str] = None,
) -> List[Dict]:
    if len(images_bytes) > 0 and "gpt" in model_name:
        return [
            {"role": "system", "content": sys_msg},
            {
                "role": "user",
                "content": [{"type": "text", "text": user_msg}]
                + [
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,{image_bytes}"},
                    }
                    for image_bytes in images_bytes
                ],
            },
        ]
    elif len(images_bytes) > 0 and "claude" in model_name:
        system_info = {"role": "system", "content": sys_msg}
        user_info = {
            "role": "user",
            "content": [{"type": "text", "text": user_msg}]
            + [
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/png",
                        "data": image_bytes,
                    },
                }
                for image_bytes in images_bytes
            ],
        }
        return [system_info, user_info]
    elif len(images_bytes) > 0:
        raise ValueError("Images are not yet supported for this model")
    else:
        return [
            {"role": "system", "content": sys_msg},
            {"role": "user", "content": user_msg},
        ]


GROQ_MODELS = [
    "llama3-8b-8192",
    "llama3-70b-8192",
    "llama-3.1-8b-instant",
    "llama-3.1-70b-versatile",
]
OLLAMA_MODELS = ["gemma2:2b", "gemma2","gemma2:27b","llama3.1","llama3.1:70b","mistral","llama3.1:405b"]
TOGETHER_MODELS = [
    "Qwen/Qwen1.5-4B-Chat",
    "meta-llama/Meta-Llama-3-8B-Instruct-Lite",
    "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo",
    "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
]
MODEL_MAP = {
    "openai": lambda x: "gpt" in x,
    "groq": lambda x: x in GROQ_MODELS,
    "claude": lambda x: "claude" in x,
    "together": lambda x: x in TOGETHER_MODELS,
    "deepmind": lambda x: "gemini" in x,
    "deepseek": lambda x: "deepseek" in x,
    "lambda": lambda x: "hermes" in x,
    "ollama": lambda x: x in OLLAMA_MODELS,
}

providers = {
    "openai": OpenAIAPIProvider,
    "groq": GroqAPIProvider,
    "claude": AnthropicAPIProvider,
    "together": TogetherAPIProvider,
    "deepmind": DeepmindAPIProvider,
    "deepseek": DeepSeekAPIProvider,
    "lambda": LambdaAPIProvider,
    "ollama": OllamaAPIProvider,
}


class LLM:
    model_name: str

    def __init__(
        self,
        model_name: str,
        temperature: Optional[float] = 0,
        max_tokens: Optional[int] = None,
        top_p: Optional[float] = None,
        frequency_penalty: Optional[float] = None,
        presence_penalty: Optional[float] = None,
        stop: Optional[str] = None,
    ):
        self.model_name = model_name
        self.temperature = temperature
        self.max_tokens = max_tokens if max_tokens else 1000
        self.top_p = top_p
        self.frequency_penalty = frequency_penalty
        self.presence_penalty = presence_penalty
        self.stop = stop

    # def save(self, system_prompt, user_prompt, response_model, response):
    #     if response_model:
    #         response = response.dict()

    def log_response(self, system_prompt: str, user_prompt: str, response: str):
        log_dir = "logs/llm/bulk"
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        now = datetime.now()
        formatted_now = now.strftime("%H_%d_%m_%Y")
        i = 0
        while os.path.exists(f"{log_dir}/log_{formatted_now}_{i}.txt"):
            i += 1
        log_content = f"System Prompt:\n```{system_prompt}```\n\nUser Prompt:```\n{user_prompt}\n\nLLM Response:\n```{response}```"
        with open(f"{log_dir}/log_{formatted_now}_{i}.txt", "w") as file:
            file.write(log_content)

    def sync_respond(
        self,
        system_prompt: str,
        user_prompt: str,
        images_bytes: List[str] = [],
        response_model: Optional[Type[BaseModel]] = None,
    ):
        messages = build_messages(
            sys_msg=system_prompt,
            user_msg=user_prompt,
            images_bytes=images_bytes,
            model_name=self.model_name,
        )
        provider_name = next(
            (k for k, v in MODEL_MAP.items() if v(self.model_name)), None
        )
        provider = (
            providers[provider_name]()
            if provider_name != "groq"
            else providers["groq"](force_structured_output=True)
        )
        if response_model:
            return provider.sync_chat_completion_with_response_model(
                messages,
                model=self.model_name,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
            )
        else:
            return provider.sync_chat_completion(
                messages,
                model=self.model_name,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
            )

    async def async_respond(
        self,
        system_prompt: str,
        user_prompt: str,
        images_bytes: List[str] = [],
        response_model: Optional[Type[BaseModel]] = None,
    ):
        messages = build_messages(
            sys_msg=system_prompt,
            user_msg=user_prompt,
            images_bytes=images_bytes,
            model_name=self.model_name,
        )
        provider_name = next(
            (k for k, v in MODEL_MAP.items() if v(self.model_name)), None
        )
        provider = (
            providers[provider_name]()
            if provider_name != "groq"
            else providers["groq"](force_structured_output=True)
        )
        if response_model:
            result = await provider.async_chat_completion_with_response_model(
                messages,
                model=self.model_name,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                response_model=response_model,
            )
            assert isinstance(result, str) or isinstance(
                result, BaseModel
            ), f"Result: {result}"
            return result
        else:
            result = await provider.async_chat_completion(
                messages,
                model=self.model_name,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
            )
            return result


if __name__ == "__main__":
    import base64

    lm = LLM("claude-3-haiku-20240307")
    image_path = "apropos/bench/crafter/crafter.png"

    def encode_image(image_path):
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode("utf-8")

    image_bytes = encode_image(image_path)
    response = lm.sync_respond(
        system_prompt="Hello",
        user_prompt="Hi",
        images_bytes=[image_bytes],
    )
    print(response)



import asyncio
import random
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Dict, List, Literal, Tuple, Union

from apropos.src.core.programs.dag import LM_DAG, DagRecord


@dataclass
class Question:
    information: Dict
    correctness: bool

    @abstractmethod
    async def compute_and_score_attempt(self, lm_dag: LM_DAG) -> Tuple[bool, DagRecord]:
        pass

    @abstractmethod
    def compute_and_score_attempt_sync(self, lm_dag: LM_DAG) -> Tuple[bool, DagRecord]:
        pass


class Benchmark:
    train: List[Question]
    test: List[Question]
    dev: List[Question]


class QABenchmark(Benchmark):
    train: List[Question]
    test: List[Question]
    dev: List[Question]

    def get_questions_by_patches(
        self,
        questions: List[Question],
        n: int,
        patches: List[str],
        sort_type: Literal["random", "first", "last"],
    ) -> List[Question]:
        questions_by_patch = {}
        alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        for i in range(0, len(questions), 100):
            if i // 100 >= len(alphabet):
                continue
            questions_by_patch[alphabet[i // 100]] = questions[i : i + 100]

        data = []
        for patch in patches:
            if patch not in questions_by_patch:
                raise ValueError(f"Patch {patch} not found in questions")
            data.extend(questions_by_patch[patch])

        if sort_type == "first":
            return data[:n]
        elif sort_type == "last":
            return data[-n:]
        elif sort_type == "random":
            random.seed(0)
            return random.sample(data, n)

    def score_dag_parsync(
        self,
        lm_dag: LM_DAG,
        split: Literal["train", "dev", "test"] = "dev",
        n: int = 50,
        verbose: bool = False,
        sort_type: str = "first",
        patches: List[str] = ["A"],
    ) -> Union[List[bool], Tuple[List[bool], List[DagRecord]]]:
        if split == "train":
            questions = self.train
        elif split == "dev":
            questions = self.dev
        elif split == "test":
            questions = self.test
        else:
            raise ValueError("Split must be one of 'train', 'dev', or 'test'")

        questions = self.get_questions_by_patches(questions, n, patches, sort_type)

        scores, records = [], []
        # Use ThreadPoolExecutor to parallelize the scoring
        with ThreadPoolExecutor() as executor:
            results = list(
                executor.map(
                    lambda q: q.compute_and_score_attempt_sync(lm_dag), questions
                )
            )

        for score, record in results:
            scores.append(score)
            records.append(record)
        if verbose:
            return scores, records
        return scores

    def get_questions(
        self,
        split: Literal["train", "dev", "test"] = "dev",
        n: int = 50,
        sort_type: str = "first",
        patches: List[str] = ["A"],
    ) -> List[Question]:
        if split == "train":
            questions = self.train
        elif split == "dev":
            questions = self.dev
        elif split == "test":
            questions = self.test
        else:
            raise ValueError("Split must be one of 'train', 'dev', or 'test'")

        return self.get_questions_by_patches(questions, n, patches, sort_type)

    def score_dag_sync(
        self,
        lm_dag: LM_DAG,
        split: Literal["train", "dev", "test"] = "dev",
        n: int = 50,
        verbose: bool = False,
        sort_type: str = "first",
        patches: List[str] = ["A"],
    ) -> Union[List[bool], Tuple[List[bool], List[DagRecord]]]:
        if split == "train":
            questions = self.train
        elif split == "dev":
            questions = self.dev
        elif split == "test":
            questions = self.test
        else:
            raise ValueError("Split must be one of 'train', 'dev', or 'test'")

        questions = self.get_questions_by_patches(questions, n, patches, sort_type)
        scores, records = [], []
        for question in questions:
            score, record = question.compute_and_score_attempt_sync(lm_dag)
            scores.append(score)
            records.append(record)

        if verbose:
            return scores, records
        return scores

    async def score_dag(
        self,
        lm_dag: LM_DAG,
        split: Literal["train", "dev", "test"] = "dev",
        n: int = 50,
        frozen_params: Dict = {"dag_record": None, "unfrozen_node_names": []},
        verbose=False,
        sort_type: Literal["random", "first", "last"] = "first",
        patches: List[str] = ["A"],
    ) -> Union[List[bool], Tuple[List[bool], List[DagRecord]]]:
        if split == "train":
            questions = self.train
        elif split == "dev":
            questions = self.dev
        elif split == "test":
            questions = self.test
        else:
            raise ValueError("Split must be one of 'train', 'dev', or 'test'")

        if n // 100 + 1 > len(patches):
            raise ValueError("Number of patches must be at least the number of batches")

        questions = self.get_questions_by_patches(questions, n, patches, sort_type)

        scores = []
        records = []
        tasks = [
            question.compute_and_score_attempt(lm_dag, frozen_params=frozen_params)
            for question in questions
        ]
        score_and_records = await asyncio.gather(*tasks)
        for score, record in score_and_records:
            scores.append(score)
            records.append(record)
        if verbose:
            return scores, records
        return scores


class GameBenchmark:
    pass



import os

os.environ["MINEDOJO_HEADLESS"] = "1"
import argparse
import numpy as np
from tqdm import tqdm
import gym
import smartplay

parser = argparse.ArgumentParser()
parser.add_argument("--llm_name", type=str, default="gpt-4", help="Name of the LLM")
parser.add_argument(
    "--env_names",
    type=str,
    default=None,
    help="Comma separated list of environments to run",
)

args = parser.parse_args()

if args.env_names is None:
    args.env_names = ",".join(smartplay.benchmark_games_v0)

LLM_name = args.llm_name

# Replace with your own LLM API.
# Note: query_model takes two arguments: 1) message in openai chat completion form (list of dictionaries),
#                                        2) an index to indicate where the message should be truncated if the length exceeds LLM context length.
from llm_api import get_query

query_model = get_query(LLM_name)


def compose_ingame_prompt(info, question, past_qa=[]):
    messages = [
        {"role": "system", "content": "You’re a player trying to play the game."}
    ]

    if len(info["manual"]) > 0:
        messages.append({"role": "system", "content": info["manual"]})

    if len(info["history"]) > 0:
        messages.append({"role": "system", "content": info["history"]})

    messages.append(
        {
            "role": "system",
            "content": "current step observation: {}".format(info["obs"]),
        }
    )

    if len(past_qa) > 0:
        for q, a in past_qa:
            messages.append({"role": "user", "content": q})
            messages.append({"role": "assistant", "content": a})

    messages.append({"role": "user", "content": question})

    return (
        messages,
        2,
    )  # This is the index of the history, we will truncate the history if it is too long for LLM


questions = [
    "What is the best action to take? Let's think step by step, ",
    "Choose the best executable action from the list of all actions. Write the exact chosen action.",
]


def run(env_name):
    normalized_scores = []
    env = gym.make("smartplay:{}-v0".format(env_name))
    env_steps = env.default_steps
    num_iter = env.default_iter

    def match_act(output):
        inds = [
            (i, output.lower().index(act.lower()))
            for i, act in enumerate(env.action_list)
            if act.lower() in output.lower()
        ]
        if len(inds) > 0:
            # return the action with smallest index
            return sorted(inds, key=lambda x: x[1])[0][0]
        else:
            # print("LLM failed with output \"{}\", taking action 0...".format(output))
            return 0

    rewards = []
    progresses = []
    for eps in tqdm(
        range(num_iter), desc="Evaluating LLM {} on {}".format(LLM_name, env_name)
    ):
        import wandb

        wandb.init(
            project="SmartPlay",
            config={
                "LLM": LLM_name,
                "env": env_name,
                "eps": eps,
                "num_iter": num_iter,
                "env_steps": env_steps,
            },
        )
        step = 0
        trajectories = []
        qa_history = []
        progress = [0]
        reward = 0
        rewards = []
        done = False

        columns = (
            ["Context", "Step", "OBS", "History", "Score", "Reward", "Total Reward"]
            + questions
            + ["Action"]
        )
        wandb_table = wandb.Table(columns=columns)

        _, info = env.reset()

        while step < env_steps:
            new_row = [
                info["manual"],
                step,
                info["obs"],
                info["history"],
                info["score"],
                reward,
                sum(rewards),
            ]
            wandb.log(
                {
                    "metric/total_reward".format(eps): sum(rewards),
                    "metric/score".format(eps): info["score"],
                    "metric/reward".format(eps): reward,
                }
            )

            if done:
                break

            qa_history = []
            for question in questions:
                prompt = compose_ingame_prompt(info, question, qa_history)
                answer = query_model(*prompt)
                qa_history.append((question, answer))
                new_row.append(answer)
                answer_act = answer

            a = match_act(answer_act)
            new_row.append(env.action_list[a])
            _, reward, done, info = env.step(a)
            rewards.append(reward)
            score = info["score"]

            step += 1
            wandb_table.add_data(*new_row)

        if not done:
            completion = 0
        else:
            completion = info["completed"]
        progresses.append(np.max(progress))
        wandb.log(
            {
                "rollout/rollout".format(eps): wandb_table,
                "final/total_reward": sum(rewards),
                "final/score": score,
                "final/normalized_score": smartplay.normalize_score(env_name, score),
                "final/completion": completion,
                "final/episodic_step": step,
                "final/eps": eps,
            }
        )
        normalized_scores.append(smartplay.normalize_score(env_name, score))
        del wandb_table
        wandb.finish()
    return np.average(normalized_scores)


score_dict = {}
for env_name in args.env_names.split(","):
    score_dict[env_name] = run(env_name)

print("Normalized scores on each task:", score_dict)
print("Capability scores of the LLM:", smartplay.analyze_capabilities(score_dict))



import copy


class HistoryTracker:
    def __init__(self, max_steps) -> None:
        self.max_steps = max_steps
        self.game_step = 0
        self.reset()

    def step(self, info) -> None:
        self.info.append(copy.copy(info))
        if len(self.info) > self.max_steps:
            self.info.pop(0)
        self.game_step += 1

    def reset(self) -> None:
        self.info = []
        self.game_step = 0

    def describe(self, game_step=None):
        if len(self.info) == 0:
            return ""
        game_step = self.game_step if game_step is None else game_step
        result = "Most recent {} steps of the player's in-game observation:\n\n".format(
            len(self.info)
        )
        for i, info in enumerate(self.info):
            result += "Player Observation Step {}:\n".format(
                game_step - len(self.info) + i
            )
            result += info["obs"] + "\n\n"
        return result.strip()

    def score(self):
        return sum([info["score"] for info in self.info])


def describe_act(action_list):
    return "List of all actions:\n" + "\n".join(
        ["{}. {}".format(i + 1, s) for i, s in enumerate(action_list)]
    )



#!/usr/bin/env python3
# type: ignore

import json
import os
import re
import subprocess
import time

SLEEP = 1.5
# set the URLs of each website, we use the demo sites as an example
os.environ["SHOPPING"] = "http://ec2-3-131-244-37.us-east-2.compute.amazonaws.com:7770"
os.environ["SHOPPING_ADMIN"] = (
    "http://ec2-3-131-244-37.us-east-2.compute.amazonaws.com:7780/admin"
)
os.environ["REDDIT"] = "http://ec2-3-131-244-37.us-east-2.compute.amazonaws.com:9999"
os.environ["GITLAB"] = "http://ec2-3-131-244-37.us-east-2.compute.amazonaws.com:8023"
os.environ["MAP"] = "http://ec2-3-131-244-37.us-east-2.compute.amazonaws.com:3000"
os.environ["WIKIPEDIA"] = (
    "http://ec2-3-131-244-37.us-east-2.compute.amazonaws.com:8888/wikipedia_en_all_maxi_2022-05/A/User:The_other_Kiwix_guy/Landing"
)
os.environ["HOMEPAGE"] = (
    "PASS"  # The home page is not currently hosted in the demo site
)
print("Done setting up URLs")

# First, run `python scripts/generate_test_data.py` to generate the config files
p = subprocess.run(["python", "scripts/generate_test_data.py"], capture_output=True)

# It will generate individual config file for each test example in config_files
config_path = "/Users/jp/Documents/GitHub/Apropos/apropos/bench/web_arena/webarena/config_files/examples/1.json"
assert os.path.exists(config_path)

# Make sure the URLs in the config files are replaced properly
with open(config_path, "r") as f:
    config = json.load(f)
    assert os.environ["SHOPPING_ADMIN"] in config["start_url"], (
        os.environ["SHOPPING_ADMIN"],
        config["start_url"],
    )

print("Done generating config files with the correct URLs")

# run bash prepare.sh to save all account cookies, this only needs to be done once
subprocess.run(["bash", "prepare.sh"])
print("Done saving account cookies")

# Init an environment
from browser_env import (
    Action,
    ActionTypes,
    ObservationMetadata,
    ScriptBrowserEnv,
    StateInfo,
    Trajectory,
    action2str,
    create_id_based_action,
    create_stop_action,
)
from evaluation_harness.evaluators import evaluator_router

# Init the environment
env = ScriptBrowserEnv(
    headless=False,
    slow_mo=100,
    observation_type="accessibility_tree",
    current_viewport_only=True,
    viewport_size={"width": 1280, "height": 720},
)

# example 156 as an example
config_file = "config_files/156.json"
# maintain a trajectory
trajectory: Trajectory = []

# set the environment for the current example
obs, info = env.reset(options={"config_file": config_file})
actree_obs = obs["text"]
print(actree_obs)

# You should see some output like this:
"""
[4] RootWebArea 'Projects · Dashboard · GitLab' focused: True
        [12] link 'Skip to content'
        [28] link 'Dashboard'
        [2266] button '' hasPopup: menu expanded: False
        [63] textbox 'Search GitLab' required: False
        [61] generic 'Use the shortcut key <kbd>/</kbd> to start a search'
        [79] link 'Create new...'
        [95] link 'Issues'
                [97] generic '13 assigned issues'
        [101] link 'Merge requests'
                [104] generic '8 merge requests'"""

# save the state info to the trajectory
state_info: StateInfo = {"observation": obs, "info": info}
trajectory.append(state_info)

# Now let's try to perform the action of clicking the "Merge request" link
# As the element ID is dynamic each time, we use regex to match the element as the demo
match = re.search(r"\[(\d+)\] link 'Merge requests'", actree_obs).group(1)
# Create the action click [ELEMENT_ID]
click_action = create_id_based_action(f"click [{match}]")
# Add the action to the trajectory
trajectory.append(click_action)

# Step and get the new observation
obs, _, terminated, _, info = env.step(click_action)
# New observation
actree_obs = obs["text"]
print(actree_obs)
time.sleep(SLEEP)

state_info = {"observation": obs, "info": info}
trajectory.append(state_info)

# Next click "assign to you"
match = re.search(r"\[(\d+)\] link 'Assigned to you", actree_obs).group(1)
click_action = create_id_based_action(f"click [{match}]")
trajectory.append(click_action)

obs, _, terminated, _, info = env.step(click_action)
actree_obs = obs["text"]
print(actree_obs)
time.sleep(SLEEP)
state_info = {"observation": obs, "info": info}
trajectory.append(state_info)

# add a stop action to mark the end of the trajectory
trajectory.append(create_stop_action(""))


# Demo evaluation
evaluator = evaluator_router(config_file)
score = evaluator(
    trajectory=trajectory,
    config_file=config_file,
    page=env.page,
    client=env.get_page_client(env.page),
)

# as we manually perform the task, the task should be judged as correct
assert score == 1.0
